# Datasets/NASA/PODAAC.py - THREDDSCatalogs and OPeNDAPURLs for NASA
# PO.DAAC datasets.
#
# Copyright (C) 2010 Jason J. Roberts
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (available in the file LICENSE.TXT)
# for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import datetime
import os

from GeoEco.Datasets import Dataset, QueryableAttribute, Grid
from GeoEco.Datasets.ArcGIS import ArcGISWorkspace, ArcGISRaster
from GeoEco.Datasets.Collections import DirectoryTree
from GeoEco.Datasets.GDAL import GDALDataset
from GeoEco.Datasets.OPeNDAP import THREDDSCatalog, OPeNDAPURL, OPeNDAPGrid
from GeoEco.Datasets.Virtual import TimeSeriesGridStack, GridSliceCollection, MaskedGrid, ClippedGrid, RotatedGlobalGrid, ClimatologicalGridCollection, DerivedGrid, CannyEdgeGrid, _CannyDescription
from GeoEco.DynamicDocString import DynamicDocString
from GeoEco.Internationalization import _
from GeoEco.Types import *


class MODISL3SSTTimeSeries(Grid):
    __doc__ = DynamicDocString()

    def __init__(self, satellite, temporalResolution, spatialResolution, geophysicalParameter, variableName=u'sst', qualityLevel=0, timeout=60, maxRetryTime=300, cacheDirectory=None):
        self.__doc__.Obj.ValidateMethodInvocation()
        
        # Set the URL, path parsing expressions, and variable name.

        queryableAttributeValues = {u'Satellite': satellite,
                                    u'SatelliteCode': satellite[0].upper(),
                                    u'TemporalResolution': temporalResolution,
                                    u'TemporalResolutionCode': {u'daily': u'DAY', u'8day': u'8D', u'monthly': u'MO', u'annual': u'YR'}[temporalResolution],
                                    u'SpatialResolution': spatialResolution,
                                    u'SpatialResolutionCode': spatialResolution[0],
                                    u'Algorithm': {u'sst': u'sst', u'nsst': u'sst', u'sst4': u'sst4'}[geophysicalParameter],
                                    u'Wavelength': {u'sst': u'11um', u'nsst': u'11um', u'sst4': u'4um'}[geophysicalParameter],
                                    u'GeophysicalParameter': geophysicalParameter}

        # Set the display name based on the data we are accessing.

        if variableName == u'sst':
            if geophysicalParameter != u'sst4':
                displayName = _(u'MODIS %(satellite)s L3 mapped %(temporalResolution)s %(spatialResolution)s thermal IR %(geophysicalParameter)s from NASA PO.DAAC') % {u'satellite': satellite[0].upper() + satellite[1:], u'temporalResolution': temporalResolution, u'spatialResolution': spatialResolution, u'geophysicalParameter': geophysicalParameter.upper()}
            else:
                displayName = _(u'MODIS %(satellite)s L3 mapped %(temporalResolution)s %(spatialResolution)s mid-IR %(geophysicalParameter)s from NASA PO.DAAC') % {u'satellite': satellite[0].upper() + satellite[1:], u'temporalResolution': temporalResolution, u'spatialResolution': spatialResolution, u'geophysicalParameter': geophysicalParameter.upper()}
        elif geophysicalParameter != u'sst4':
            displayName = _(u'MODIS %(satellite)s L3 mapped %(temporalResolution)s %(spatialResolution)s thermal IR %(geophysicalParameter)s quality level from NASA PO.DAAC') % {u'satellite': satellite[0].upper() + satellite[1:], u'temporalResolution': temporalResolution, u'spatialResolution': spatialResolution, u'geophysicalParameter': geophysicalParameter.upper()}
        else:
            displayName = _(u'MODIS %(satellite)s L3 mapped %(temporalResolution)s %(spatialResolution)s mid-IR %(geophysicalParameter)s quality level from NASA PO.DAAC') % {u'satellite': satellite[0].upper() + satellite[1:], u'temporalResolution': temporalResolution, u'spatialResolution': spatialResolution, u'geophysicalParameter': geophysicalParameter.upper()}


        # Construct a THREDDSCatalog for the variable requested by the
        # caller and wrap it in a TimeSeriesGridStack to create a 3D
        # grid with dimensions tyx.

        url = 'http://opendap.jpl.nasa.gov/opendap/allData/modis/L3/%s/%s/v2014.0/%s/%s/catalog.xml' % (satellite, queryableAttributeValues[u'Wavelength'], spatialResolution, temporalResolution)
        if temporalResolution == u'daily':
            pathParsingExpressions=[r'(?P<Year>\d\d\d\d)',
                                    r'(?P<DayOfYear>\d\d\d)',
                                    r'%s(?P<Year>\d\d\d\d)(?P<DayOfYear>\d\d\d)(?P<EndDate>\d{0})\.L3m_%s_%s_%s_%s\.nc' % (queryableAttributeValues[u'SatelliteCode'], queryableAttributeValues[u'TemporalResolutionCode'], geophysicalParameter, queryableAttributeValues[u'Algorithm'], queryableAttributeValues[u'SpatialResolution'])]
        else:
            pathParsingExpressions=[r'(?P<Year>\d\d\d\d)',
                                    r'%s(?P<Year>\d\d\d\d)(?P<DayOfYear>\d\d\d)(?P<EndDate>\d\d\d\d\d\d\d)\.L3m_%s_%s_%s_%s\.nc' % (queryableAttributeValues[u'SatelliteCode'], queryableAttributeValues[u'TemporalResolutionCode'], geophysicalParameter, queryableAttributeValues[u'Algorithm'], queryableAttributeValues[u'SpatialResolution'])]

        if variableName == u'sst':
            variableName = queryableAttributeValues[u'Algorithm']
        else:
            variableName = u'qual_' + queryableAttributeValues[u'Algorithm']

        grid = TimeSeriesGridStack(THREDDSCatalog(url=url,
                                                  pathParsingExpressions=pathParsingExpressions,
                                                  displayName=displayName,
                                                  timeout=timeout,
                                                  maxRetryTime=maxRetryTime,
                                                  queryableAttributes=self._GetQueryableAttributesForTimeSlices(),
                                                  queryableAttributeValues=queryableAttributeValues,
                                                  lazyPropertyValues=self._GetLazyPropertyValuesForTimeSlices(variableName, spatialResolution, temporalResolution, geophysicalParameter),
                                                  cacheDirectory=cacheDirectory),
                                   reportProgress=False)

        # If the caller requested a specific quality level, create a
        # THREDDSCatalog for the pixel quality variable and use it to
        # mask the 3D grid.

        if qualityLevel is not None and qualityLevel in [0, 1]:
            grid = MaskedGrid(grid,
                              masks=[TimeSeriesGridStack(THREDDSCatalog(url=url,
                                                                        pathParsingExpressions=pathParsingExpressions,
                                                                        displayName=_(u'quality level'),
                                                                        timeout=timeout,
                                                                        maxRetryTime=maxRetryTime,
                                                                        queryableAttributes=self._GetQueryableAttributesForTimeSlices(),
                                                                        queryableAttributeValues=queryableAttributeValues,
                                                                        lazyPropertyValues=self._GetLazyPropertyValuesForTimeSlices(u'qual_' + queryableAttributeValues[u'Algorithm'], spatialResolution, temporalResolution, geophysicalParameter),
                                                                        cacheDirectory=cacheDirectory),
                                                         reportProgress=False)],
                              operators=[u'>'],
                              values=[qualityLevel])

        # Assign our _WrappedGrid instance variable.

        self._WrappedGrid = grid

    def __getattribute__(self, name):
        if name.startswith('__') or name in ['_WrappedGrid', '_GetTimeCoordsFromQueryableAttributeValues', '_GetQueryableAttributesForTimeSlices', '_GetLazyPropertyValuesForTimeSlices', 'CreateArcGISRasters', 'CreateClimatologicalArcGISRasters', 'InterpolateAtArcGISPoints', 'CreateCayulaCornillonFrontsAsArcGISRasters', 'CannyEdgesAsArcGISRasters']:
            return object.__getattribute__(self, name)
        return object.__getattribute__(object.__getattribute__(self, '_WrappedGrid'), name)

    def __setattr__(self, name, value):
        if name.startswith('__') or name in ['_WrappedGrid', '_GetTimeCoordsFromQueryableAttributeValues', '_GetQueryableAttributesForTimeSlices', '_GetLazyPropertyValuesForTimeSlices', 'CreateArcGISRasters', 'CreateClimatologicalArcGISRasters', 'InterpolateAtArcGISPoints', 'CreateCayulaCornillonFrontsAsArcGISRasters', 'CannyEdgesAsArcGISRasters']:
            object.__setattr__(self, name, value)
        else:
            setattr(object.__getattribute__(self, '_WrappedGrid'), name, value)

    @classmethod
    def _GetQueryableAttributesForTimeSlices(cls):

        # First define a function used to calculate the value of the
        # EndDate queryable attribute from the value of the DateTime
        # queryable attribute.
        
        def _GetEndDate(selfOrDict, startDate):

            # Get the temporal resolution.
            
            temporalResolution = None
            if isinstance(selfOrDict, dict):
                if u'TemporalResolution' in selfOrDict:
                    temporalResolution = selfOrDict[u'TemporalResolution']
            else:
                temporalResolution = selfOrDict.GetQueryableAttributeValue(u'TemporalResolution')

            # Calculate the end date string from the temporal resolution.

            if temporalResolution is not None:
                temporalResolution = temporalResolution.lower()
                
                if temporalResolution == u'daily':
                    return ''
                
                if temporalResolution == u'8day':
                    return unicode((startDate + datetime.timedelta(days=7)).strftime('%Y%j'))
                
                if temporalResolution == u'monthly':
                    if startDate.month == 12:
                        return unicode((datetime.datetime(startDate.year + 1, 1, 1) - datetime.timedelta(days=1)).strftime('%Y%j'))
                    return unicode((datetime.datetime(startDate.year, startDate.month + 1, 1) - datetime.timedelta(days=1)).strftime('%Y%j'))
                
                if temporalResolution == u'annual':
                    return unicode((datetime.datetime(startDate.year + 1, 1, 1) - datetime.timedelta(days=1)).strftime('%Y%j'))
                    
            return None

        # Now return a tuple of the queryable attributes. The EndDate
        # queryable attribute makes use of the function.

        return (QueryableAttribute(u'Satellite', _(u'Satellite'), UnicodeStringTypeMetadata(allowedValues=[u'aqua', u'terra'], makeLowercase=True)),
                QueryableAttribute(u'SatelliteCode', _(u'Satellite abbreviation code'), UnicodeStringTypeMetadata(allowedValues=[u'a', u't'], makeLowercase=True)),
                QueryableAttribute(u'TemporalResolution', _(u'Temporal resolution'), UnicodeStringTypeMetadata(allowedValues=[u'daily', u'8day', u'monthly', u'annual'], makeLowercase=True)),
                QueryableAttribute(u'TemporalResolutionCode', _(u'Temporal resolution abbreviation code'), UnicodeStringTypeMetadata(allowedValues=[u'DAY', u'8D', u'MO', u'YR'], makeLowercase=True)),
                QueryableAttribute(u'SpatialResolution', _(u'Spatial resolution'), UnicodeStringTypeMetadata(allowedValues=[u'4km', u'9km'], makeLowercase=True)),
                QueryableAttribute(u'SpatialResolutionCode', _(u'Spatial resolution abbreviation code'), UnicodeStringTypeMetadata(allowedValues=[u'4', u'9'], makeLowercase=True)),
                QueryableAttribute(u'Algorithm', _(u'Algorithm'), UnicodeStringTypeMetadata(allowedValues=[u'sst', u'sst4'], makeLowercase=True)),
                QueryableAttribute(u'Wavelength', _(u'Primary wavelength'), UnicodeStringTypeMetadata(allowedValues=[u'11um', u'4um'], makeLowercase=True)),
                QueryableAttribute(u'GeophysicalParameter', _(u'Geophysical parameter'), UnicodeStringTypeMetadata(allowedValues=[u'sst', u'nsst', u'sst4'], makeLowercase=True)),
                QueryableAttribute(u'DateTime', _(u'Start date'), DateTimeTypeMetadata()),
                QueryableAttribute(u'EndDate', _(u'End date string'), UnicodeStringTypeMetadata(mustMatchRegEx=ur'\d\d\d\d\d\d\d\d'), derivedFromAttr=u'DateTime', derivedValueFunc=_GetEndDate),
                QueryableAttribute(u'VariableName', _(u'Variable'), UnicodeStringTypeMetadata(allowedValues=[u'sst', u'qual_sst', u'sst4', u'qual_sst4'], makeLowercase=True)))

    @classmethod
    def _GetLazyPropertyValuesForTimeSlices(cls, variableName, spatialResolution, temporalResolution, geophysicalParameter):
        import numpy
        return {'VariableTypes': ['Grid'],
                'VariableNames': [variableName],
                'SpatialReference': Dataset.ConvertSpatialReference('proj4', '+proj=latlong +ellps=WGS84 +datum=WGS84 +no_defs', 'obj'),
                'Dimensions': 'yx',
                'Shape': {'4km': (4320, 8640), '9km': (2160, 4320)}[spatialResolution],
                'CoordDependencies': (None, None),
                'CoordIncrements': {'4km': (180./4320,360./8640), '9km': (180./2160,360./4320)}[spatialResolution],
                'TIncrement': {'daily': 1, '8day': 8, 'monthly': 1, u'annual': 1}[temporalResolution],
                'TIncrementUnit': {'daily': 'day', '8day': 'day', 'monthly': 'month', u'annual': 'year'}[temporalResolution],
                'TSemiRegularity': {'daily': None, '8day': 'annual', 'monthly': None, u'annual': None}[temporalResolution],
                'TCountPerSemiRegularPeriod': {'daily': None, '8day': 46, 'monthly': None, u'annual': None}[temporalResolution],
                'TCornerCoordType': 'min',
                'TOffsetFromParsedTime': {'sst': None, 'nsst': -0.5, 'sst4': -0.5}[geophysicalParameter],       # Nighttime datasets start 1/2 days prior to the date and time parsed from the file (i.e. they start at 12:00:00 on the previous date, rather than 00:00:00 on the parsed date)
                'CornerCoords': {'4km': (-90 + 180./4320/2, -180 + 360./8640/2), '9km': (-90 + 180./2160/2, -180 + 360./4320/2)}[spatialResolution],
                'PhysicalDimensions': 'yx',
                'PhysicalDimensionsFlipped': (True, False),
                'UnscaledDataType': {'sst': 'int16', 'qual_sst': 'int16'}[variableName.strip('4')],
                'UnscaledNoDataValue': {'sst': -32767, 'qual_sst': -1}[variableName.strip('4')],
                'ScaledDataType': {'sst': 'float32', 'qual_sst': None}[variableName.strip('4')],
                'ScaledNoDataValue': {'sst': numpy.cast['float32'](0.00499999989*-32767), 'qual_sst': None}[variableName.strip('4')],
                'ScalingFunction': {'sst': lambda data: numpy.cast['float32'](0.00499999989*data), 'qual_sst': None}[variableName.strip('4')],
                'UnscalingFunction': {'sst': lambda data: numpy.cast['iint16'](numpy.round(data/0.00499999989)), 'qual_sst': None}[variableName.strip('4')]}

    @classmethod
    def _GetTimeCoordsFromQueryableAttributeValues(cls, queryableAttributeValues):
        try:
            if u'DateTime' in queryableAttributeValues and isinstance(queryableAttributeValues[u'DateTime'], datetime.datetime):
                startDate = queryableAttributeValues[u'DateTime']
                temporalResolution = None

                if u'GeophysicalParameter' in queryableAttributeValues and isinstance(queryableAttributeValues[u'GeophysicalParameter'], basestring) and queryableAttributeValues[u'GeophysicalParameter'].lower() in [u'nsst', u'sst4']:
                    tOffset = datetime.timedelta(-0.5)
                else:
                    tOffset = datetime.timedelta(0)
                
                if u'TemporalResolution' in queryableAttributeValues and isinstance(queryableAttributeValues[u'TemporalResolution'], basestring) and queryableAttributeValues[u'TemporalResolution'] in [u'daily', '8day', 'monthly', 'annual']:
                    temporalResolution = queryableAttributeValues[u'TemporalResolution']
                elif u'TemporalResolutionCode' in queryableAttributeValues and isinstance(queryableAttributeValues[u'TemporalResolutionCode'], basestring) and queryableAttributeValues[u'TemporalResolutionCode'] in [u'DAY', '8D', 'MO', 'YR']:
                    temporalResolution = queryableAttributeValues[u'TemporalResolutionCode']
                else:
                    return [startDate + tOffset, None, None]
                    
                if temporalResolution in [u'daily', u'DAY']:
                    return [startDate + tOffset, startDate + datetime.timedelta(0.5) + tOffset, startDate + datetime.timedelta(1) + tOffset - datetime.timedelta(seconds=1)]
                
                if temporalResolution in [u'8day', u'8D']:
                    if startDate.month == 12 and startDate.day >= 24:
                        endDate = datetime.datetime(startDate.year + 1, 1, 1)
                        return [startDate + tOffset, startDate + (endDate - startDate) / 2 + tOffset, endDate + tOffset - datetime.timedelta(seconds=1)]
                    else:
                        return [startDate + tOffset, startDate + datetime.timedelta(4) + tOffset, startDate + datetime.timedelta(8) + tOffset - datetime.timedelta(seconds=1)]
                    
                if temporalResolution in [u'monthly', u'MO']:
                    if startDate.month == 12:
                        endDate = datetime.datetime(startDate.year + 1, 1, 1)
                    else:
                        endDate = datetime.datetime(startDate.year, startDate.month + 1, 1)
                    return [startDate + tOffset, startDate + (endDate - startDate) / 2 + tOffset, endDate + tOffset - datetime.timedelta(seconds=1)]

                if temporalResolution in [u'annual', u'YR']:
                    return [startDate + tOffset, datetime.datetime(startDate.year, 7, 1) + tOffset, datetime.datetime(startDate.year + 1, 1, 1) + tOffset - datetime.timedelta(seconds=1)]

            return [None, None, None]
        except:
            return [None, None, None]

    @classmethod
    def _RotateAndClip(cls, grid, rotationOffset, spatialExtent, startDate, endDate):
        if rotationOffset is not None:
            grid = RotatedGlobalGrid(grid, rotationOffset, u'Map units')

        xMin, yMin, xMax, yMax = None, None, None, None
        if spatialExtent is not None:
            from GeoEco.Types import EnvelopeTypeMetadata
            xMin, yMin, xMax, yMax = EnvelopeTypeMetadata.ParseFromArcGISString(spatialExtent)

        if spatialExtent is not None or startDate is not None or endDate is not None:
            if startDate is not None:
                startDate = datetime.datetime(startDate.year, startDate.month, startDate.day, 0, 0, 0)
            if endDate is not None:
                endDate = datetime.datetime(endDate.year, endDate.month, endDate.day, 23, 59, 59)
                
            grid = ClippedGrid(grid, u'Map coordinates', xMin=xMin, xMax=xMax, yMin=yMin, yMax=yMax, tMin=startDate, tMax=endDate)
            
        return grid

    @classmethod
    def CreateArcGISRasters(cls, satellite, temporalResolution, spatialResolution, geophysicalParameter, variableName,
                            outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(Satellite)s', u'%(Wavelength)s', u'%(SpatialResolution)s', u'%(TemporalResolution)s', u'%%Y', u'%(SatelliteCode).1s%%Y%%j%(EndDate)s.L3m_%(TemporalResolutionCode).3s_%(GeophysicalParameter)s_%(SpatialResolutionCode).1s_%(VariableName)s.img'], rasterCatalog=None,
                            qualityLevel=0,
                            rotationOffset=None, spatialExtent=None, startDate=None, endDate=None, 
                            timeout=60, maxRetryTime=300, cacheDirectory=None,
                            useUnscaledData=False, calculateStatistics=True, buildRAT=False, buildPyramids=False):
        cls.__doc__.Obj.ValidateMethodInvocation()
        grid = MODISL3SSTTimeSeries(satellite, temporalResolution, spatialResolution, geophysicalParameter, variableName, qualityLevel, timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)
            workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=cls._GetQueryableAttributesForTimeSlices())
            workspace.ImportDatasets(GridSliceCollection(grid).QueryDatasets(), mode, useUnscaledData=useUnscaledData, calculateStatistics=calculateStatistics, buildRAT=buildRAT, buildPyramids=buildPyramids)
            if rasterCatalog is not None:
                workspace.ToRasterCatalog(rasterCatalog, grid.GetSpatialReference(u'ArcGIS'), tQACoordType=u'min', tCoordFunction=cls._GetTimeCoordsFromQueryableAttributeValues, overwriteExisting=True)
        finally:
            grid.Close()
        return outputWorkspace

    @classmethod
    def CreateClimatologicalArcGISRasters(cls, satellite, temporalResolution, spatialResolution, geophysicalParameter, variableName,
                                          statistic, binType,
                                          outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(Satellite)s', u'%(Wavelength)s', u'%(SpatialResolution)s', u'%(TemporalResolution)s', u'%(ClimatologyBinType)s_Climatology', u'%(Satellite)s_%(GeophysicalParameter)s_%(VariableName)s_%(ClimatologyBinName)s_%(Statistic)s.img'],
                                          qualityLevel=0,
                                          binDuration=1, startDayOfYear=1,
                                          rotationOffset=None, spatialExtent=None, startDate=None, endDate=None,
                                          timeout=60, maxRetryTime=300, cacheDirectory=None,
                                          calculateStatistics=True, buildPyramids=False):
        cls.__doc__.Obj.ValidateMethodInvocation()
        grid = MODISL3SSTTimeSeries(satellite, temporalResolution, spatialResolution, geophysicalParameter, variableName, qualityLevel, timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)
            collection = ClimatologicalGridCollection(grid, statistic, binType, binDuration, startDayOfYear, reportProgress=True)
            workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=tuple(collection.GetAllQueryableAttributes()))
            workspace.ImportDatasets(collection.QueryDatasets(), mode, calculateStatistics=calculateStatistics, buildPyramids=buildPyramids)
        finally:
            grid.Close()
        return outputWorkspace

    @classmethod
    def InterpolateAtArcGISPoints(cls, satellite, temporalResolution, spatialResolution, geophysicalParameter, variableName,
                                  points, valueField, tField, method=u'Nearest', where=None, noDataValue=None,
                                  qualityLevel=0,
                                  timeout=60, maxRetryTime=300, cacheDirectory=None,
                                  orderByFields=None, numBlocksToCacheInMemory=256, xBlockSize=16, yBlockSize=16, tBlockSize=3):
        cls.__doc__.Obj.ValidateMethodInvocation()
        
        grid = MODISL3SSTTimeSeries(satellite, temporalResolution, spatialResolution, geophysicalParameter, variableName, qualityLevel, timeout, maxRetryTime, cacheDirectory)

        try:
            if orderByFields is not None:
                orderBy = u', '.join(map(lambda f: f + u' ASC', orderByFields))
            else:
                from GeoEco.ArcGIS import GeoprocessorManager
                if GeoprocessorManager.GetArcGISMajorVersion() > 9 or GeoprocessorManager.GetArcGISMinorVersion() >= 2:
                    orderBy = tField + u' ASC'
                else:
                    orderBy = None
            
            from GeoEco.Datasets.ArcGIS import ArcGISTable
            from GeoEco.SpatialAnalysis.Interpolation import Interpolator
            Interpolator.InterpolateGridsValuesForTableOfPoints([grid], ArcGISTable(points), [valueField], tField=tField, where=where, orderBy=orderBy, method=method, noDataValue=noDataValue, gridsWrap=True, numBlocksToCacheInMemory=numBlocksToCacheInMemory, xBlockSize=xBlockSize, yBlockSize=yBlockSize, tBlockSize=tBlockSize)

        finally:
            grid.Close()
        
        return points

    @classmethod
    def CreateCayulaCornillonFrontsAsArcGISRasters(cls, satellite, temporalResolution, spatialResolution, geophysicalParameter, minPopMeanDifference,
                                                   outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(Satellite)s', u'%(Wavelength)s_fronts', u'%(SpatialResolution)s', u'%(TemporalResolution)s', u'%%Y', '%(Satellite)s_%(TemporalResolution)s_%%Y%%j_%(GeophysicalParameter)s_%(ImageType)s.img'],
                                                   medianFilterWindowSize=3, histogramWindowSize=20, histogramWindowStride=1, minPropNonMaskedCells=0.65, minPopProp=0.25, minTheta=0.76, minSinglePopCohesion=0.88, minGlobalPopCohesion=0.90, threads=1,
                                                   fillHoles=20, thin=True, minSize=10,
                                                   qualityLevel=1,
                                                   rotationOffset=None, spatialExtent=None, startDate=None, endDate=None,
                                                   timeout=60, maxRetryTime=300, cacheDirectory=None,
                                                   calculateStatistics=True, buildRAT=False, buildPyramids=False,
                                                   outputCandidateCounts=False, outputFrontCounts=False, outputWindowStatusCodes=False, outputWindowStatusValues=False):
        cls.__doc__.Obj.ValidateMethodInvocation()

        # Construct a MODISL3SSTTimeSeries instance and rotate
        # and clip it as requested.

        grid = MODISL3SSTTimeSeries(satellite, temporalResolution, spatialResolution, geophysicalParameter, u'sst', qualityLevel, timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)

            # Construct a CayulaCornillonFrontsInGrid collection using
            # the requested parameters.

            from GeoEco.OceanographicAnalysis.Fronts import CayulaCornillonFrontsInGrid

            collection = CayulaCornillonFrontsInGrid(grid,
                                                     wrapEdges=grid.Shape[-1] == MODISL3SSTTimeSeries._GetLazyPropertyValuesForTimeSlices(u'sst', spatialResolution, temporalResolution, geophysicalParameter)['Shape'][-1],
                                                     medianFilterWindowSize=medianFilterWindowSize,
                                                     histogramWindowSize=histogramWindowSize,
                                                     histogramWindowStride=histogramWindowStride,
                                                     minPropNonMaskedCells=minPropNonMaskedCells,
                                                     minPopProp=minPopProp,
                                                     minPopMeanDifference=minPopMeanDifference,
                                                     minTheta=minTheta,
                                                     minSinglePopCohesion=minSinglePopCohesion,
                                                     minGlobalPopCohesion=minGlobalPopCohesion,
                                                     threads=threads,
                                                     fillHoles=fillHoles,
                                                     thin=thin,
                                                     minSize=minSize)
            try:
                try:
                    # Construct an ArcGISWorkspace instance and import
                    # time slices from the CayulaCornillonFrontsInGrid
                    # instance.

                    expression = "ImageType = 'floc'"
                    if outputCandidateCounts:
                        expression += " OR ImageType = 'ccnt'"
                    if outputFrontCounts:
                        expression += " OR ImageType = 'fcnt'"
                    if outputWindowStatusCodes:
                        expression += " OR ImageType = 'wsco'"
                    if outputWindowStatusValues:
                        expression += " OR ImageType = 'wsvl'"

                    workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=tuple(collection.GetAllQueryableAttributes()))
                    workspace.ImportDatasets(collection.QueryDatasets(expression), mode, calculateStatistics=calculateStatistics, buildRAT=buildRAT, buildPyramids=buildPyramids)

                # If we caught an exception, log it now. Then delete
                # the collection object, to ensure its memory is
                # freed. Normally we don't bother doing this but the
                # memory it holds can be substantial.
                
                except:
                    from GeoEco.Logging import Logger
                    Logger.LogExceptionAsError()
            finally:
                del collection
        finally:
            grid.Close()

        # Return successfully.

        return outputWorkspace

    @classmethod
    def CannyEdgesAsArcGISRasters(cls, satellite, temporalResolution, spatialResolution, geophysicalParameter,
                                  outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(Satellite)s', u'%(Wavelength)s_Canny_fronts', u'%(SpatialResolution)s', u'%(TemporalResolution)s', u'%%Y', '%(Satellite)s_%(TemporalResolution)s_%%Y%%j_%(GeophysicalParameter)s_Canny_fronts.img'],
                                  highThreshold=None, lowThreshold=None, sigma=1.4, minSize=4,
                                  qualityLevel=1,
                                  rotationOffset=None, spatialExtent=None, startDate=None, endDate=None,
                                  timeout=600, maxRetryTime=None, cacheDirectory=None,
                                  calculateStatistics=True, buildRAT=False, buildPyramids=False):
        cls.__doc__.Obj.ValidateMethodInvocation()
        grid = MODISL3SSTTimeSeries(satellite, temporalResolution, spatialResolution, geophysicalParameter, u'sst', qualityLevel, timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)
            grid = CannyEdgeGrid(grid, highThreshold, lowThreshold, sigma, minSize)
            workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=cls._GetQueryableAttributesForTimeSlices())
            workspace.ImportDatasets(GridSliceCollection(grid).QueryDatasets(), mode, calculateStatistics=calculateStatistics, buildRAT=buildRAT, buildPyramids=buildPyramids)
        finally:
            grid.Close()
        return outputWorkspace


class _GHRSSTLevel4THREDDSCatalog(THREDDSCatalog):
    __doc__ = DynamicDocString()

    def __init__(self, gdsVersion=1, fileVersionString=None, customPathParsingExpressions=None, displayName=None, timeout=60, maxRetryTime=120, cacheDirectory=None, cacheCatalogsOnDisk=False):
        if gdsVersion == 1:
            if customPathParsingExpressions is None:
                customPathParsingExpressions = [r'(?P<AreaCode>[^\-]+)',
                                                r'(?P<RDACCode>[^\-]+)',
                                                r'(?P<ProductCode2>[^\-]+)',
                                                r'(?P<Year>\d\d\d\d)',
                                                r'(?P<DayOfYear>\d\d\d)',
                                                r'(?P<Year>\d\d\d\d)(?P<Month>\d\d)(?P<Day>\d\d)-(?P<RDACCode>[^\-]+)-L4(?P<ProductType>[^\-]+)-(?P<AreaCode2>[^\-]+)-v(?P<GDSVersion>\d\d)-fv(?P<FileVersion>[^\-]+)-(?P<ProductCode>[^\-]+)(?:\.nc|\.nc\.bz2|\.nc\.gz)']
            super(_GHRSSTLevel4THREDDSCatalog, self).__init__('http://podaac-opendap.jpl.nasa.gov/opendap/allData/ghrsst/data/L4/catalog.xml',
                                                              customPathParsingExpressions,
                                                              displayName=displayName,
                                                              timeout=timeout,
                                                              maxRetryTime=maxRetryTime,
                                                              queryableAttributes=(QueryableAttribute(u'RDACCode', _(u'RDAC code'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'ProductType', _(u'Product type'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'AreaCode', _(u'Area code'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'AreaCode2', _(u'Area code 2'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'GDSVersion', _(u'GDS version'), IntegerTypeMetadata()),
                                                                                   QueryableAttribute(u'FileVersion', _(u'File version'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'ProductCode', _(u'Product code'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'ProductCode2', _(u'Product code 2'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'DateTime', _(u'Start date'), DateTimeTypeMetadata())),
                                                              cacheDirectory=cacheDirectory,
                                                              cacheCatalogsOnDisk=cacheCatalogsOnDisk)
        elif gdsVersion == 2:
            if customPathParsingExpressions is None:
                customPathParsingExpressions = [r'(?P<AreaCode>[^\-]+)',
                                                r'(?P<RDACCode>[^\-]+)',
                                                r'(?P<ProductCode2>[^\-]+)',
                                                '(v' + fileVersionString.strip('0').strip('.') + '|v' + fileVersionString.lstrip('0') + ')',
                                                r'(?P<Year>\d\d\d\d)',
                                                r'(?P<DayOfYear>\d\d\d)',
                                                r'(?P<Year>\d\d\d\d)(?P<Month>\d\d)(?P<Day>\d\d)(?P<Hour>\d\d)(?P<Minute>\d\d)(?P<Second>\d\d)-(?P<RDACCode>[^\-]+)-L4_GHRSST-(?P<ProductType>[^\-]+)-(?P<ProductCode>[^\-]+)-(?P<AreaCode2>[^\-]+)-v(?P<GDSVersion>\d\d)\.0-fv(?P<FileVersion>[^\-]+)(?:\.nc|\.nc\.bz2|\.nc\.gz)']
            super(_GHRSSTLevel4THREDDSCatalog, self).__init__('http://podaac-opendap.jpl.nasa.gov/opendap/allData/ghrsst/data/GDS2/L4/catalog.xml',
                                                              customPathParsingExpressions,
                                                              displayName=displayName,
                                                              timeout=timeout,
                                                              maxRetryTime=maxRetryTime,
                                                              queryableAttributes=(QueryableAttribute(u'RDACCode', _(u'RDAC code'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'ProductType', _(u'Product type'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'AreaCode', _(u'Area code'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'AreaCode2', _(u'Area code 2'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'GDSVersion', _(u'GDS version'), IntegerTypeMetadata()),
                                                                                   QueryableAttribute(u'FileVersion', _(u'File version'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'ProductCode', _(u'Product code'), UnicodeStringTypeMetadata()),
                                                                                   QueryableAttribute(u'ProductCode2', _(u'Product code 2'), UnicodeStringTypeMetadata()), #derivedFromAttr=u'ProductCode', derivedValueFunc=lambda obj, value: value),
                                                                                   QueryableAttribute(u'DateTime', _(u'Start date'), DateTimeTypeMetadata())),
                                                              cacheDirectory=cacheDirectory,
                                                              cacheCatalogsOnDisk=cacheCatalogsOnDisk)
        else:
            raise NotImplementedError(_(u'Programming error in this tool: unknown gdsVersion %(gdsVersion)r. Please contact the MGET development team for assistance.') % {u'gdsVersion': gdsVersion})

    def _ConstructOPeNDAPURL(self, url, timeout, maxRetryTime, queryableAttributeValues, cacheDirectory):
        return _GHRSSTLevel4OPeNDAPURL(url,
                                       timeout=timeout,
                                       maxRetryTime=maxRetryTime,
                                       parentCollection=self,
                                       queryableAttributeValues=queryableAttributeValues,
                                       lazyPropertyValues={'VariableTypes': ['Grid', 'Grid', 'Grid', 'Grid'],
                                                           'VariableNames': ['analysed_sst', 'analysis_error', 'mask', 'sea_ice_fraction']},
                                       cacheDirectory=cacheDirectory)


class _GHRSSTLevel4OPeNDAPURL(OPeNDAPURL):
    __doc__ = DynamicDocString()

    def _ConstructOPeNDAPGrid(self, variableName, variableType):
        return _GHRSSTLevel4OPeNDAPGrid(self, variableName, variableType)


class _GHRSSTLevel4OPeNDAPGrid(OPeNDAPGrid):
    __doc__ = DynamicDocString()

    def __init__(self, opendapURLObj, variableName, variableType):
        self.__doc__.Obj.ValidateMethodInvocation()

        # Perform additional validation.
        
        if variableName not in ['analysed_sst', 'analysis_error', 'mask', 'sea_ice_fraction']:
            raise ValueError(_(u'This tool does not support the "%(variableName)s" variable. The supported variables are "analysed_sst", "analysis_error", "mask", and "sea_ice_fraction".') % {u'variableName': variableName})

        # Initialize the base class.

        if opendapURLObj.GetQueryableAttributeValue('GDSVersion') == 1:
            tCornerCoordType = 'min'        # In GDS 1.0, GHRSST did not include the hour, minute, and second in the file name. We assumed that the averaging window started at midnight on that day.
        elif opendapURLObj.GetQueryableAttributeValue('GDSVersion') == 2:
            tCornerCoordType = 'center'     # In GDS 2.0, GHRSST started including the hour, minute, and second in the file name and said for L4 products that the "indicative time" for L4 products is the "nominal time of analysis" and "All times should be given in UTC and should be chosen to best represent the observation time for this dataset." Given this, and the time stamps I observed in the products, the time stamp is the "center" of the averaging window.
        else:
            raise ValueError(_(u'Programming error in this tool. For %(dn)s, the GDSVersion was %(GDSVersion)r, which is an unknown value. Please contact the MGET development team for assistance.') % {u'dn': opendapURLObj.DisplayName, u'GDSVersion': opendapURLObj.GetQueryableAttributeValue('GDSVersion')})
        
        super(_GHRSSTLevel4OPeNDAPGrid, self).__init__(opendapURLObj,
                                                       variableName,
                                                       variableType,
                                                       lazyPropertyValues={'SpatialReference': Dataset.ConvertSpatialReference('proj4', '+proj=latlong +ellps=WGS84 +datum=WGS84 +no_defs', 'obj'),
                                                                           'Dimensions': u'tyx',
                                                                           'CoordDependencies': (None, None, None),
                                                                           'TIncrement': 1,
                                                                           'TIncrementUnit': 'day',
                                                                           'TSemiRegularity': None,
                                                                           'TCountPerSemiRegularPeriod': None,
                                                                           'TCornerCoordType': tCornerCoordType,
                                                                           'PhysicalDimensions': u'tyx'})

    def _GetLazyPropertyPhysicalValue(self, name):

        # If this is a lazy property that is common to all time slices
        # of this GHRSST product, check our cache for the value.

        if name in ['Shape', 'CoordIncrements', 'CornerCoords', 'PhysicalDimensionsFlipped', 'UnscaledDataType', 'ScaledDataType', 'UnscaledNoDataValue', 'ScaledNoDataValue', 'ScalingFunction', 'UnscalingFunction'] and self._CommonLPPVIsCached(name):
            return self._GetCachedCommonLPPV(name)

        # If the caller is asking for the CoordIncrements or
        # CornerCoords, calculate them from the lon and lat OPeNDAP
        # variables, which give the center coordinates of the cells.

        self.ParentCollection._Open()

        if name in ['CoordIncrements', 'CornerCoords', 'PhysicalDimensionsFlipped']:
            lon = self.ParentCollection._PydapDataset['lon']
            lat = self.ParentCollection._PydapDataset['lat']

            yIncrement = abs((lat[-1] - lat[0]) / (lat.shape[0] - 1))

            if os.path.basename(self.ParentCollection.URL).find('-GLOB-') >= 0:      # If it is a global dataset, assume 360 degrees longitude and calculate the x increment at full precision.
                xIncrement = 360. / self.Shape[-1]
            elif lon[0] > lon[-1]:
                xIncrement = abs((lon[-1] + 360. - lon[0]) / (lon.shape[0] - 1))     # For datasets such as ABOM-L4HRfnd-AUS-RAMSSA_09km that cross the 180th meridian, use a 0-to-360 coordinate system
            else:
                xIncrement = abs((lon[-1] - lon[0]) / (lon.shape[0] - 1))
                    
            if abs(yIncrement - xIncrement) < 0.000001:     # If x and y increments are almost exactly the same, force them to be exactly the same so that the cells are perfectly square
                yIncrement = xIncrement

            self._SetCachedCommonLPPV('CoordIncrements', (1., yIncrement, xIncrement))

            if lat[-1] > lat[0]:
                self._SetCachedCommonLPPV('CornerCoords', (datetime.datetime(1980, 1, 1), lat[0], lon[0]))      # Use fake time coordinate; the true time coordinates are parsed from the file name.
                self._SetCachedCommonLPPV('PhysicalDimensionsFlipped', (False, False, False))
            else:
                self._SetCachedCommonLPPV('CornerCoords', (datetime.datetime(1980, 1, 1), lat[-1], lon[0]))      # Use fake time coordinate; the true time coordinates are parsed from the file name.
                self._SetCachedCommonLPPV('PhysicalDimensionsFlipped', (False, True, False))

            return self._GetCachedCommonLPPV(name)

        # If the caller is asking for the scaled data type, scaling
        # function, or unscaling function, determine these from the
        # data type and add_offset and scale_factor attributes of this
        # OPeNDAP variable.

        if name in ['ScaledDataType', 'ScalingFunction', 'UnscalingFunction']:
            v = self._GetOPeNDAPVariable()

            if self.GetLazyPropertyValue('UnscaledDataType').startswith('float') or not hasattr(v, 'scale_factor') or not hasattr(v, 'add_offset'):
                self._SetCachedCommonLPPV('ScaledDataType', None)
                self._SetCachedCommonLPPV('ScalingFunction', None)
                self._SetCachedCommonLPPV('UnscalingFunction', None)
            else:
                import numpy
                self._SetCachedCommonLPPV('ScaledDataType', u'float32')
                self._SetCachedCommonLPPV('ScalingFunction', lambda data: numpy.cast['float32'](v.scale_factor*data + v.add_offset))
                self._SetCachedCommonLPPV('UnscalingFunction', lambda data: numpy.cast[str(self.GetLazyPropertyValue('UnscaledDataType'))](numpy.round((data - v.add_offset)/v.scale_factor)))

            return self._GetCachedCommonLPPV(name)

        # If the caller is asking for the UnscaledNoDataValue, get it
        # from the _FillValue attribute of this OPeNDAP variable.

        if name == 'UnscaledNoDataValue':
            v = self._GetOPeNDAPVariable()

            if not hasattr(v, '_FillValue'):
                self._SetCachedCommonLPPV('UnscaledNoDataValue', None)
            elif not self.GetLazyPropertyValue('UnscaledDataType').startswith('float'):
                self._SetCachedCommonLPPV('UnscaledNoDataValue', int(v._FillValue))
            else:
                self._SetCachedCommonLPPV('UnscaledNoDataValue', v._FillValue)

            return self._GetCachedCommonLPPV(name)

        # If the caller is asking for the ScaledNoDataValue, calculate
        # it.
        
        if name == 'ScaledNoDataValue':
            scalingFunction = self.GetLazyPropertyValue('ScalingFunction')
            unscaledNoDataValue = self.GetLazyPropertyValue('UnscaledNoDataValue')

            if scalingFunction is None or unscaledNoDataValue is None:
                self._SetCachedCommonLPPV('ScaledNoDataValue', None)
            else:
                self._SetCachedCommonLPPV('ScaledNoDataValue', scalingFunction(unscaledNoDataValue))

            return self._GetCachedCommonLPPV(name)

        # Otherwise let the base class get this property.

        self._SetCachedCommonLPPV(name, super(_GHRSSTLevel4OPeNDAPGrid, self)._GetLazyPropertyPhysicalValue(name))
        return self._GetCachedCommonLPPV(name)

    _CommonLPPVCache = {}

    def _CommonLPPVIsCached(self, name):
        key = self._GetCommonLPPVCacheKey(name)
        return key in _GHRSSTLevel4OPeNDAPGrid._CommonLPPVCache

    def _GetCachedCommonLPPV(self, name):
        key = self._GetCommonLPPVCacheKey(name)
        if key in _GHRSSTLevel4OPeNDAPGrid._CommonLPPVCache:
            return _GHRSSTLevel4OPeNDAPGrid._CommonLPPVCache[key]
        return None

    def _SetCachedCommonLPPV(self, name, value):
        key = self._GetCommonLPPVCacheKey(name)
        _GHRSSTLevel4OPeNDAPGrid._CommonLPPVCache[key] = value

    def _GetCommonLPPVCacheKey(self, name):
        return ','.join([name, self.VariableName] + [str(self.GetQueryableAttributeValue(qa)) for qa in [u'RDACCode', u'ProductType', u'AreaCode', u'AreaCode2', u'GDSVersion', u'FileVersion', u'ProductCode', u'ProductCode2']])


class GHRSSTLevel4(Grid):
    __doc__ = DynamicDocString()

    def __init__(self, variableName, productName, rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2=None, applyMask=False, customPathParsingExpressions=None, convertToCelsius=True, timeout=60, maxRetryTime=300, cacheDirectory=None, cacheCatalogsOnDisk=False):
        self.__doc__.Obj.ValidateMethodInvocation()

        # Construct a _GHRSSTLevel4THREDDSCatalog and wrap it in a
        # TimeSeriesGridStack to create a 3D grid with dimensions tyx.

        if productCode2 is None:
            productCode2 = productCode

        variableNames = [variableName]
        rdacCodes = [rdacCode]
        productTypes = [productType]
        areaCodes = [areaCode]
        gdsVersions = [gdsVersion]
        fileVersions = [fileVersion]
        productCode2s = [productCode2]

        # if rdacCode == 'REMSS' and areaCode == 'GLOB' and productCode == 'mw_ir_rt_OI':
        #     datasetSpecificExpression = 'AND (Year > 2008 OR Year = 2008 AND DayOfYear >= 199)'     # Hack to speed up querying for mw_ir_rt_OI, which is stored in same directory as mw_ir_OI but starts at a much later date.
        # elif rdacCode == 'JPL' and areaCode == 'GLOB' and productCode == 'MUR' and fileVersion == '03':
        #     fileVersions = ['03', '04']
        #     datasetSpecificExpression = "AND (FileVersion = '03' AND (Year < 2010 OR Year = 2010 AND DayOfYear <= 183) OR FileVersion = '04' AND (Year > 2010 OR Year = 2010 AND DayOfYear > 183))"     # Hack to handle revision of MUR algorithm that did not result in reprocessing of entire dataset
        # else:
        #     datasetSpecificExpression = ''
        #
        # As of May 2016 (and probably sometime earlier) it appears that the
        # hacks above are no longer necessary. The REMSS products have been
        # replaced by GDS 2.0 versions and the GDS 1.0 MUR product has been
        # reprocessed back to the beginning of the time series with file
        # version 04. So I am removing these hacks.
        #
        # In these queries, I am omitting ProductCode and AreaCode2. They are
        # redundant with ProductCode2 and AreaCode. By omitting them, some
        # queries will not need to descend all the way into day-of-year
        # directories on the THREDDS server. If they are not omitted, queries
        # must descend there so that they can be obtained from the GHRSST file
        # name, which makes certain queries much slower.

        datasetSpecificExpression = ''

        grid = TimeSeriesGridStack(_GHRSSTLevel4THREDDSCatalog(gdsVersion, fileVersion, customPathParsingExpressions, _('GHRSST L4 product %s hosted by NASA PO.DAAC') % productName, timeout, maxRetryTime, cacheDirectory, cacheCatalogsOnDisk),
                                   expression = "VariableName IN (%s) AND RDACCode IN (%s) AND ProductType IN (%s) AND AreaCode IN (%s) AND GDSVersion IN (%s) AND FileVersion IN (%s) AND ProductCode2 IN (%s) %s" % \
                                                (', '.join(["'" + s + "'" for s in variableNames]),
                                                 ', '.join(["'" + s + "'" for s in rdacCodes]),
                                                 ', '.join(["'" + s + "'" for s in productTypes]),
                                                 ', '.join(["'" + s + "'" for s in areaCodes]),
                                                 ', '.join([str(i) for i in gdsVersions]),
                                                 ', '.join(["'" + s + "'" for s in fileVersions]),
                                                 ', '.join(["'" + s + "'" for s in productCode2s]),
                                                 datasetSpecificExpression),
                                   reportProgress=False)

        # If the caller requested that we apply the mask, wrap the
        # grid.

        if applyMask:
            grid = MaskedGrid(grid,
                              masks=[TimeSeriesGridStack(_GHRSSTLevel4THREDDSCatalog(gdsVersion, fileVersion, customPathParsingExpressions, u'the mask variable', timeout, maxRetryTime, cacheDirectory, cacheCatalogsOnDisk),
                                                         expression = "VariableName = 'mask' AND RDACCode IN (%s) AND ProductType IN (%s) AND AreaCode IN (%s) AND GDSVersion IN (%s) AND FileVersion IN (%s) AND ProductCode2 IN (%s) %s" % \
                                                         (', '.join(["'" + s + "'" for s in rdacCodes]),
                                                          ', '.join(["'" + s + "'" for s in productTypes]),
                                                          ', '.join(["'" + s + "'" for s in areaCodes]),
                                                          ', '.join([str(i) for i in gdsVersions]),
                                                          ', '.join(["'" + s + "'" for s in fileVersions]),
                                                          ', '.join(["'" + s + "'" for s in productCode2s]),
                                                          datasetSpecificExpression),
                                   reportProgress=False)],
                              operators=[u'!='],
                              values=[1])

        # If the caller requested that we convert Kelvin temperatures
        # to Celsius, wrap the grid.

        if convertToCelsius:
            qav = {}
            for qa in grid.GetAllQueryableAttributes():
                qav[qa.Name] = grid.GetQueryableAttributeValue(qa.Name)
            
            grid = DerivedGrid([grid],
                               'self._Grids[0].Data.__getitem__(tuple(sliceList)) - 273.15',
                               grid.DisplayName,
                               grid.DataType,
                               grid.NoDataValue - 273.15,
                               queryableAttributes=tuple(grid.GetAllQueryableAttributes()),
                               queryableAttributeValues=qav)

        # Assign our _WrappedGrid instance variable.

        self._WrappedGrid = grid

    def __getattribute__(self, name):
        if name.startswith('__') or name in ['_WrappedGrid', '_PODAAC_Products', '_RotateAndClip', 'CreateArcGISRasters', 'CreateClimatologicalArcGISRasters', 'InterpolateAtArcGISPoints', 'CreateCayulaCornillonFrontsAsArcGISRasters', 'CannyEdgesAsArcGISRasters']:
            return object.__getattribute__(self, name)
        return object.__getattribute__(object.__getattribute__(self, '_WrappedGrid'), name)

    def __setattr__(self, name, value):
        if name.startswith('__') or name in ['_WrappedGrid', '_PODAAC_Products', '_RotateAndClip', 'CreateArcGISRasters', 'CreateClimatologicalArcGISRasters', 'InterpolateAtArcGISPoints', 'CreateCayulaCornillonFrontsAsArcGISRasters', 'CannyEdgesAsArcGISRasters']:
            object.__setattr__(self, name, value)
        else:
            setattr(object.__getattribute__(self, '_WrappedGrid'), name, value)

    # Define a custom path parsing expression for the GOS products, which do
    # not follow the convention of having a subdirectory level with the GHRSST
    # "file version" that matches the one at the end of the file names. With
    # the GOS products, the directory level is 'v2.0' while the file name
    # contains 'fv1.0'.

    _GOS_CustomPathParsingExpressions = [r'(?P<AreaCode>[^\-]+)',
                                         r'(?P<RDACCode>[^\-]+)',
                                         r'(?P<ProductCode2>[^\-]+)',
                                         'v2.0',
                                         r'(?P<Year>\d\d\d\d)',
                                         r'(?P<DayOfYear>\d\d\d)',
                                         r'(?P<Year>\d\d\d\d)(?P<Month>\d\d)(?P<Day>\d\d)(?P<Hour>\d\d)(?P<Minute>\d\d)(?P<Second>\d\d)-(?P<RDACCode>[^\-]+)-L4_GHRSST-(?P<ProductType>[^\-]+)-(?P<ProductCode>[^\-]+)-(?P<AreaCode2>[^\-]+)-v(?P<GDSVersion>\d\d)\.0-fv(?P<FileVersion>[^\-]+)(?:\.nc|\.nc\.bz2|\.nc\.gz)']

    # In the lists below the items are [rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions]

    _PODAAC_Products = {# GHRSST Data Processing Specification (GDS) 1.0 products
                        u'ABOM-L4HRfnd-AUS-RAMSSA_09km': ['ABOM', 'HRfnd', 'AUS', 'AUS', 1, '01_0', 'RAMSSA_09km', 'RAMSSA_09km', True, None],
                        u'ABOM-L4LRfnd-GLOB-GAMSSA_28km': ['ABOM', 'LRfnd', 'GLOB', 'GLOB', 1, '01_0', 'GAMSSA_28km', 'GAMSSA_28km', True, None],
                        u'DMI-L4UHfnd-NSEABALTIC-DMI_OI': ['DMI', 'UHfnd', 'NSEABALTIC', 'NSEABALTIC', 1, '01', 'DMI_OI', 'DMI_OI', False, None],
                        u'EUR-L4UHRfnd-MED-ODYSSEA': ['EUR', 'UHRfnd', 'MED', 'MED', 1, '01', 'ODYSSEA', 'ODYSSEA', False, None],
                        u'EUR-L4UHRfnd-NWE-ODYSSEA': ['EUR', 'UHRfnd', 'NWE', 'NWE', 1, '01', 'ODYSSEA', 'ODYSSEA', False, None],
                        u'JPL-L4UHfnd-GLOB-MUR': ['JPL', 'UHfnd', 'GLOB', 'GLOB', 1, '04', 'MUR', 'MUR', False, None],
                        u'JPL_OUROCEAN-L4UHfnd-GLOB-G1SST': ['JPL_OUROCEAN', 'UHfnd', 'GLOB', 'GLOB', 1, '01_0', 'G1SST', 'G1SST', False, None],
                        u'NAVO-L4HR1m-GLOB-K10_SST': ['NAVO', 'HR1m', 'GLOB', 'GLOB', 1, '01_0', 'K10_SST', 'K10_SST', False, None],
                        u'NCDC-L4LRblend-GLOB-AVHRR_AMSR_OI': ['NCDC', 'LRblend', 'GLOB', 'GLOB', 1, '02_0', 'AVHRR_AMSR_OI', 'AVHRR_AMSR_OI', False, None],
                        u'NCDC-L4LRblend-GLOB-AVHRR_OI': ['NCDC', 'LRblend', 'GLOB', 'GLOB', 1, '02_0', 'AVHRR_OI', 'AVHRR_OI', False, None],
                        u'UKMO-L4HRfnd-GLOB-OSTIA': ['UKMO', 'HRfnd', 'GLOB', 'GLOB', 1, '02', 'OSTIA', 'OSTIA', False, None],

                        # GHRSST Data Processing Specification (GDS) 2.0 products
                        u'OISST_HR_NRT-GOS-L4-BLK-v2.0': ['GOS', 'SSTfnd', 'BLACKSEA', 'BLK', 2, '01.0', 'OISST_HR_NRT', 'OISST_HR', True, _GOS_CustomPathParsingExpressions],
                        u'OISST_UHR_NRT-GOS-L4-BLK-v2.0': ['GOS', 'SSTfnd', 'BLACKSEA', 'BLK', 2, '01.0', 'OISST_UHR_NRT', 'OISST_UHR', True, _GOS_CustomPathParsingExpressions],
                        u'CMC0.1deg-CMC-L4-GLOB-v3.0': ['CMC', 'SSTfnd', 'GLOB', 'GLOB', 2, '03.0', 'CMC0.1deg', 'CMC0.1deg', True, None],
                        u'CMC0.2deg-CMC-L4-GLOB-v2.0': ['CMC', 'SSTfnd', 'GLOB', 'GLOB', 2, '02.0', 'CMC0.2deg', 'CMC0.2deg', True, None],
                        u'DMI_OI-DMI-L4-GLOB-v1.0': ['DMI', 'SSTfnd', 'GLOB', 'GLOB', 2, '01.0', 'DMI_OI', 'DMI_OI', False, None],
                        u'MUR-JPL-L4-GLOB-v4.1': ['JPL', 'SSTfnd', 'GLOB', 'GLOB', 2, '04.1', 'MUR', 'MUR', False, None],
                        u'AVHRR_OI-NCEI-L4-GLOB-v2.0': ['NCEI', 'SSTblend', 'GLOB', 'GLOB', 2, '02.0', 'AVHRR_OI', 'AVHRR_OI', False, None],
                        u'Geo_Polar_Blended-OSPO-L4-GLOB-v1.0': ['OSPO', 'SSTfnd', 'GLOB', 'GLOB', 2, '01.0', 'Geo_Polar_Blended', 'Geo_Polar_Blended', False, None],
                        u'Geo_Polar_Blended_Night-OSPO-L4-GLOB-v1.0': ['OSPO', 'SSTfnd', 'GLOB', 'GLOB', 2, '01.0', 'Geo_Polar_Blended_Night', 'Geo_Polar_Blended_Night', False, None],
                        u'MW_OI-REMSS-L4-GLOB-v4.0': ['REMSS', 'SSTsubskin', 'GLOB', 'fusion', 2, '01.0', 'MW', 'mw_OI', False, [r'(?P<AreaCode>[^\-]+)',
                                                                                                                               r'(?P<RDACCode>[^\-]+)',
                                                                                                                               r'(?P<ProductCode2>[^\-]+)',
                                                                                                                               'v4.0',
                                                                                                                               r'(?P<Year>\d\d\d\d)',
                                                                                                                               r'(?P<DayOfYear>\d\d\d)',
                                                                                                                               r'(?P<Year>\d\d\d\d)(?P<Month>\d\d)(?P<Day>\d\d)(?P<Hour>\d\d)(?P<Minute>\d\d)(?P<Second>\d\d)-(?P<RDACCode>[^\-]+)-L4_GHRSST-(?P<ProductType>[^\-]+)-(?P<ProductCode>[^\-]+)-mw\.(?P<AreaCode2>[^\-\.]+)\.\d\d\d\d\.\d\d\d\.[^\-]+-v(?P<GDSVersion>\d\d)\.0-fv(?P<FileVersion>[^\-]+)(?:\.nc|\.nc\.bz2|\.nc\.gz)']],
                        u'MW_IR_OI-REMSS-L4-GLOB-v4.0': ['REMSS', 'SSTsubskin', 'GLOB', 'fusion', 2, '01.0', 'MW_IR', 'mw_ir_OI', False, [r'(?P<AreaCode>[^\-]+)',
                                                                                                                                        r'(?P<RDACCode>[^\-]+)',
                                                                                                                                        r'(?P<ProductCode2>[^\-]+)',
                                                                                                                                        'v4.0',
                                                                                                                                        r'(?P<Year>\d\d\d\d)',
                                                                                                                                        r'(?P<DayOfYear>\d\d\d)',
                                                                                                                                        r'(?P<Year>\d\d\d\d)(?P<Month>\d\d)(?P<Day>\d\d)(?P<Hour>\d\d)(?P<Minute>\d\d)(?P<Second>\d\d)-(?P<RDACCode>[^\-]+)-L4_GHRSST-(?P<ProductType>[^\-]+)-(?P<ProductCode>[^\-]+)-mw_ir\.(?P<AreaCode2>[^\-\.]+)\.\d\d\d\d\.\d\d\d\.[^\-]+-v(?P<GDSVersion>\d\d)\.0-fv(?P<FileVersion>[^\-]+)(?:\.nc|\.nc\.bz2|\.nc\.gz)']],
                        u'OSTIA-UKMO-L4-GLOB-v2.0': ['UKMO', 'SSTfnd', 'GLOB', 'GLOB', 2, '02.0', 'OSTIA', 'OSTIA', False, None],
                        u'OISST_HR_NRT-GOS-L4-MED-v2.0': ['GOS', 'SSTfnd', 'MEDITERRANEAN', 'MED', 2, '01.0', 'OISST_HR_NRT', 'OISST_HR', True, _GOS_CustomPathParsingExpressions],
                        u'OISST_UHR_NRT-GOS-L4-MED-v2.0': ['GOS', 'SSTfnd', 'MEDITERRANEAN', 'MED', 2, '01.0', 'OISST_UHR_NRT', 'OISST_UHR', True, _GOS_CustomPathParsingExpressions],
                        u'REMO_OI_SST_5km-UFRJ-L4-SAMERICA-v1.0': ['UFRJ', 'SSTfnd', 'SAMERICA', 'SAMERICA', 2, '01.0', 'REMO_OI_SST_5km', 'REMO_OI_SST_5km', False, None],
                       }

    @classmethod
    def _RotateAndClip(cls, grid, rotationOffset, spatialExtent, startDate, endDate):
        if rotationOffset is not None:
            grid = RotatedGlobalGrid(grid, rotationOffset, u'Map units')

        xMin, yMin, xMax, yMax = None, None, None, None
        if spatialExtent is not None:
            from GeoEco.Types import EnvelopeTypeMetadata
            xMin, yMin, xMax, yMax = EnvelopeTypeMetadata.ParseFromArcGISString(spatialExtent)

        if spatialExtent is not None or startDate is not None or endDate is not None:
            if startDate is not None:
                startDate = datetime.datetime(startDate.year, startDate.month, startDate.day, 0, 0, 0)
            if endDate is not None:
                endDate = datetime.datetime(endDate.year, endDate.month, endDate.day, 23, 59, 59)
                
            grid = ClippedGrid(grid, u'Map coordinates', xMin=xMin, xMax=xMax, yMin=yMin, yMax=yMax, tMin=startDate, tMax=endDate)
            
        return grid

    @classmethod
    def CreateArcGISRasters(cls, product, variableName,
                            outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(AreaCode)s', u'%(RDACCode)s', u'%(ProductCode2)s', u'%(VariableName)s', u'%%Y', u'%%Y%%m%%d%%H%%M%%S-%(RDACCode)s-L4%(ProductType)s-%(AreaCode2)s-v%(GDSVersion)02i-fv%(FileVersion)s-%(ProductCode)s-%(VariableName)s.img'], rasterCatalog=None,
                            rotationOffset=None, spatialExtent=None, startDate=None, endDate=None, 
                            timeout=60, maxRetryTime=300, cacheDirectory=None,
                            convertToCelsius=True, useUnscaledData=False, calculateStatistics=True, buildRAT=False, buildPyramids=False):
        cls.__doc__.Obj.ValidateMethodInvocation()
        [rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions] = cls._PODAAC_Products[product]
        grid = cls(variableName, product, rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions, convertToCelsius and variableName == u'analysed_sst' and not useUnscaledData, timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)
            workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=tuple(grid.GetAllQueryableAttributes() + [QueryableAttribute(u'DateTime', _(u'Date'), DateTimeTypeMetadata())]))
            workspace.ImportDatasets(GridSliceCollection(grid, tQACoordType=grid.GetLazyPropertyValue('TCornerCoordType')).QueryDatasets(), mode, useUnscaledData=useUnscaledData, calculateStatistics=calculateStatistics, buildRAT=buildRAT, buildPyramids=buildPyramids)
            if rasterCatalog is not None:
                workspace.ToRasterCatalog(rasterCatalog, grid.GetSpatialReference(u'ArcGIS'), tQACoordType=grid.GetLazyPropertyValue('TCornerCoordType'), tCoordFunction=lambda qav: [qav[u'DateTime'], qav[u'DateTime'] + datetime.timedelta(0.5), qav[u'DateTime'] + datetime.timedelta(1) - datetime.timedelta(seconds=1)], overwriteExisting=True)
        finally:
            grid.Close()
        return outputWorkspace

    @classmethod
    def CreateClimatologicalArcGISRasters(cls, product, variableName,
                                          statistic, binType,
                                          outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(AreaCode)s', u'%(RDACCode)s', u'%(ProductCode2)s', u'%(VariableName)s', u'%(ClimatologyBinType)s_Climatology', u'%(RDACCode)s-L4%(ProductType)s-%(AreaCode2)s-v%(GDSVersion)02i-fv%(FileVersion)s-%(ProductCode)s-%(VariableName)s-%(ClimatologyBinName)s-%(Statistic)s.img'],
                                          binDuration=1, startDayOfYear=1,
                                          rotationOffset=None, spatialExtent=None, startDate=None, endDate=None,
                                          timeout=60, maxRetryTime=300, cacheDirectory=None,
                                          convertToCelsius=True, calculateStatistics=True, buildPyramids=False):
        cls.__doc__.Obj.ValidateMethodInvocation()
        [rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions] = cls._PODAAC_Products[product]
        grid = cls(variableName, product, rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions, convertToCelsius and variableName == u'analysed_sst', timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)
            collection = ClimatologicalGridCollection(grid, statistic, binType, binDuration, startDayOfYear, reportProgress=True)
            workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=tuple(collection.GetAllQueryableAttributes()))
            workspace.ImportDatasets(collection.QueryDatasets(), mode, calculateStatistics=calculateStatistics, buildPyramids=buildPyramids)
        finally:
            grid.Close()
        return outputWorkspace

    @classmethod
    def InterpolateAtArcGISPoints(cls, product, variableName,
                                  points, valueField, tField, method=u'Nearest', where=None, noDataValue=None, convertToCelsius=True,
                                  timeout=60, maxRetryTime=300, cacheDirectory=None,
                                  orderByFields=None, numBlocksToCacheInMemory=256, xBlockSize=64, yBlockSize=64, tBlockSize=1):
        cls.__doc__.Obj.ValidateMethodInvocation()
        [rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions] = cls._PODAAC_Products[product]
        grid = cls(variableName, product, rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions, convertToCelsius and variableName == u'analysed_sst', timeout, maxRetryTime, cacheDirectory)
        try:
            if orderByFields is not None:
                orderBy = u', '.join(map(lambda f: f + u' ASC', orderByFields))
            else:
                from GeoEco.ArcGIS import GeoprocessorManager
                if GeoprocessorManager.GetArcGISMajorVersion() > 9 or GeoprocessorManager.GetArcGISMinorVersion() >= 2:
                    orderBy = tField + u' ASC'
                else:
                    orderBy = None
            from GeoEco.Datasets.ArcGIS import ArcGISTable
            from GeoEco.SpatialAnalysis.Interpolation import Interpolator
            Interpolator.InterpolateGridsValuesForTableOfPoints([grid], ArcGISTable(points), [valueField], tField=tField, where=where, orderBy=orderBy, method=method, noDataValue=noDataValue, gridsWrap=True, numBlocksToCacheInMemory=numBlocksToCacheInMemory, xBlockSize=xBlockSize, yBlockSize=yBlockSize, tBlockSize=tBlockSize)
        finally:
            grid.Close()
        return points

    @classmethod
    def CreateCayulaCornillonFrontsAsArcGISRasters(cls, product, minPopMeanDifference,
                                                   outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(AreaCode)s', u'%(RDACCode)s', u'%(ProductCode2)s', u'%(VariableName)s', u'%%Y', '%%Y%%m%%d%%H%%M%%S-%(RDACCode)s-L4%(ProductType)s-%(AreaCode2)s-v%(GDSVersion)02i-fv%(FileVersion)s-%(ProductCode)s-%(ImageType)s.img'],
                                                   medianFilterWindowSize=3, histogramWindowSize=32, histogramWindowStride=1, minPropNonMaskedCells=0.65, minPopProp=0.25, minTheta=0.76, minSinglePopCohesion=0.90, minGlobalPopCohesion=0.92, threads=1,
                                                   fillHoles=20, thin=True, minSize=10,
                                                   rotationOffset=None, spatialExtent=None, startDate=None, endDate=None,
                                                   timeout=60, maxRetryTime=300, cacheDirectory=None,
                                                   calculateStatistics=True, buildRAT=False, buildPyramids=False,
                                                   outputCandidateCounts=False, outputFrontCounts=False, outputWindowStatusCodes=False, outputWindowStatusValues=False):
        cls.__doc__.Obj.ValidateMethodInvocation()

        # Construct a GHRSSTLevel4 instance and rotate and clip it as
        # requested.

        [rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions] = cls._PODAAC_Products[product]
        grid = cls(u'analysed_sst', product, rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions, False, timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)

            # Construct a CayulaCornillonFrontsInGrid collection using
            # the requested parameters.

            from GeoEco.OceanographicAnalysis.Fronts import CayulaCornillonFrontsInGrid

            collection = CayulaCornillonFrontsInGrid(grid,
                                                     tQACoordType=grid.GetLazyPropertyValue('TCornerCoordType'),
                                                     wrapEdges='-GLOB-' in product and spatialExtent is None,
                                                     medianFilterWindowSize=medianFilterWindowSize,
                                                     histogramWindowSize=histogramWindowSize,
                                                     histogramWindowStride=histogramWindowStride,
                                                     minPropNonMaskedCells=minPropNonMaskedCells,
                                                     minPopProp=minPopProp,
                                                     minPopMeanDifference=minPopMeanDifference,
                                                     minTheta=minTheta,
                                                     minSinglePopCohesion=minSinglePopCohesion,
                                                     minGlobalPopCohesion=minGlobalPopCohesion,
                                                     threads=threads,
                                                     fillHoles=fillHoles,
                                                     thin=thin,
                                                     minSize=minSize)
            try:
                try:
                    # Construct an ArcGISWorkspace instance and import
                    # time slices from the CayulaCornillonFrontsInGrid
                    # instance.

                    expression = "ImageType = 'floc'"
                    if outputCandidateCounts:
                        expression += " OR ImageType = 'ccnt'"
                    if outputFrontCounts:
                        expression += " OR ImageType = 'fcnt'"
                    if outputWindowStatusCodes:
                        expression += " OR ImageType = 'wsco'"
                    if outputWindowStatusValues:
                        expression += " OR ImageType = 'wsvl'"

                    workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=tuple(collection.GetAllQueryableAttributes()))
                    workspace.ImportDatasets(collection.QueryDatasets(expression), mode, calculateStatistics=calculateStatistics, buildRAT=buildRAT, buildPyramids=buildPyramids)

                # If we caught an exception, log it now. Then delete
                # the collection object, to ensure its memory is
                # freed. Normally we don't bother doing this but the
                # memory it holds can be substantial.
                
                except:
                    from GeoEco.Logging import Logger
                    Logger.LogExceptionAsError()
            finally:
                del collection
        finally:
            grid.Close()

        # Return successfully.

        return outputWorkspace

    @classmethod
    def CannyEdgesAsArcGISRasters(cls, product,
                                  outputWorkspace, mode=u'add', rasterNameExpressions=[u'%(AreaCode)s', u'%(RDACCode)s', u'%(ProductCode2)s', u'%(VariableName)s', u'%%Y', '%%Y%%m%%d%%H%%M%%S-%(RDACCode)s-L4%(ProductType)s-%(AreaCode2)s-v%(GDSVersion)02i-fv%(FileVersion)s-%(ProductCode)s-Canny_fronts.img'],
                                  highThreshold=None, lowThreshold=None, sigma=1.4, minSize=4,
                                  rotationOffset=None, spatialExtent=None, startDate=None, endDate=None,
                                  timeout=600, maxRetryTime=None, cacheDirectory=None,
                                  calculateStatistics=True, buildRAT=False, buildPyramids=False):
        cls.__doc__.Obj.ValidateMethodInvocation()
        [rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions] = cls._PODAAC_Products[product]
        grid = cls(u'analysed_sst', product, rdacCode, productType, areaCode, areaCode2, gdsVersion, fileVersion, productCode, productCode2, applyMask, customPathParsingExpressions, False, timeout, maxRetryTime, cacheDirectory)
        try:
            grid = cls._RotateAndClip(grid, rotationOffset, spatialExtent, startDate, endDate)
            grid = CannyEdgeGrid(grid, highThreshold, lowThreshold, sigma, minSize)
            workspace = ArcGISWorkspace(outputWorkspace, ArcGISRaster, pathCreationExpressions=rasterNameExpressions, cacheTree=True, queryableAttributes=tuple(grid.GetAllQueryableAttributes() + [QueryableAttribute(u'DateTime', _(u'Date'), DateTimeTypeMetadata())]))
            workspace.ImportDatasets(GridSliceCollection(grid, tQACoordType=grid.GetLazyPropertyValue('TCornerCoordType')).QueryDatasets(), mode, calculateStatistics=calculateStatistics, buildRAT=buildRAT, buildPyramids=buildPyramids)
        finally:
            grid.Close()
        return outputWorkspace


class _CCMPWindsL35THREDDSCatalog(THREDDSCatalog):
    __doc__ = DynamicDocString()

    def __init__(self, temporalResolution, displayName=None, timeout=60, maxRetryTime=120, cacheDirectory=None):

        queryableAttributes = (QueryableAttribute(u'TemporalResolution', _(u'Temporal resolution'), UnicodeStringTypeMetadata(allowedValues=[u'5day', u'monthly'], makeLowercase=True)),
                               QueryableAttribute(u'TemporalResolutionCode', _(u'Temporal resolution abbreviation code'), UnicodeStringTypeMetadata(allowedValues=[u'pentad', u'month'], makeLowercase=True)),
                               QueryableAttribute(u'ProductVersion', _(u'Product version number string'), UnicodeStringTypeMetadata(minLength=2, maxLength=2, makeLowercase=True)),
                               QueryableAttribute(u'ProductLevel', _(u'Product level designation string'), UnicodeStringTypeMetadata(minLength=2, maxLength=2, makeLowercase=True)),
                               QueryableAttribute(u'ProductStream', _(u'Product stream'), UnicodeStringTypeMetadata(allowedValues=[u'flk', u'llk'], makeLowercase=True)),
                               QueryableAttribute(u'DateTime', _(u'Start date'), DateTimeTypeMetadata()),
                               QueryableAttribute(u'VariableName', _(u'Variable'), UnicodeStringTypeMetadata(allowedValues=[u'uwnd', u'vwnd', u'upstr', u'vpstr', u'wspd', u'nobs'], makeLowercase=True)))

        queryableAttributeValues = {u'TemporalResolution': temporalResolution,
                                    u'TemporalResolutionCode': {u'5day': u'pentad', u'monthly': u'month'}[temporalResolution]}

        pathParsingExpressions = [r'%s' % {u'5day': 'pentad', u'monthly': 'monthly'}[temporalResolution],
                                  r'(?P<ProductStream>[a-zA-Z0-9]{3})',
                                  r'(?P<Year>\d\d\d\d)',
                                  r'(?P<Month>\d\d)',
                                  r'%s_(?P<Year>\d\d\d\d)(?P<Month>\d\d)(?P<Day>\d\d)_v(?P<ProductVersion>[0-9]{2})l(?P<ProductLevel>[0-9]{2})(?P<ProductStream>[a-zA-Z0-9]{3})\.nc\.gz' % queryableAttributeValues[u'TemporalResolutionCode']]

        if temporalResolution == u'monthly':        # There is no month level in the directory structure for the monthly product, so delete it.
            del pathParsingExpressions[3]

        super(_CCMPWindsL35THREDDSCatalog, self).__init__('http://podaac-opendap.jpl.nasa.gov/opendap/allData/ccmp/L3.5a/catalog.xml',
                                                          pathParsingExpressions=pathParsingExpressions,
                                                          displayName=displayName,
                                                          timeout=timeout,
                                                          maxRetryTime=maxRetryTime,
                                                          queryableAttributes=queryableAttributes,
                                                          queryableAttributeValues=queryableAttributeValues,
                                                          cacheDirectory=cacheDirectory,
                                                          cacheCatalogsOnDisk=True)

    def _ConstructOPeNDAPURL(self, url, timeout, maxRetryTime, queryableAttributeValues, cacheDirectory):
        return _CCMPWindsL35OPeNDAPURL(url,
                                       timeout=timeout,
                                       maxRetryTime=maxRetryTime,
                                       parentCollection=self,
                                       queryableAttributeValues=queryableAttributeValues,
                                       lazyPropertyValues={'VariableTypes': ['Grid', 'Grid', 'Grid', 'Grid', 'Grid', 'Grid'],
                                                           'VariableNames': ['uwnd', 'vwnd', 'upstr', 'vpstr', 'wspd', 'nobs']},
                                       cacheDirectory=cacheDirectory)


class _CCMPWindsL35OPeNDAPURL(OPeNDAPURL):
    __doc__ = DynamicDocString()

    def _ConstructOPeNDAPGrid(self, variableName, variableType):
        return _CCMPWindsL35OPeNDAPGrid(self, variableName, variableType)


class _CCMPWindsL35OPeNDAPGrid(OPeNDAPGrid):
    __doc__ = DynamicDocString()

    def __init__(self, opendapURLObj, variableName, variableType):
        self.__doc__.Obj.ValidateMethodInvocation()

        import numpy
        
        temporalResolution = opendapURLObj.GetQueryableAttributeValue(u'TemporalResolution')

        if variableName == 'nobs':
            scaledDataType = 'int16'
        else:
            scaledDataType = 'float32'
            
        super(_CCMPWindsL35OPeNDAPGrid, self).__init__(opendapURLObj,
                                                       variableName,
                                                       variableType,
                                                       lazyPropertyValues = {'VariableTypes': ['Grid'],
                                                                             'VariableNames': [variableName],
                                                                             'SpatialReference': Dataset.ConvertSpatialReference('proj4', '+proj=latlong +ellps=WGS84 +datum=WGS84 +no_defs', 'obj'),
                                                                             'Dimensions': 'tyx',
                                                                             'Shape': (1, 628, 1440),
                                                                             'CoordDependencies': (None, None, None),
                                                                             'CoordIncrements': ({u'5day': 5, u'monthly': 1}[temporalResolution], 0.25, 0.25),
                                                                             'TIncrement': {u'5day': 5, u'monthly': 1}[temporalResolution],                        # Note that having this property is redundant with CoordIncrements[0] but is currently required by TimeSeriesGridStack
                                                                             'TIncrementUnit': {u'5day': 'day', u'monthly': 'month'}[temporalResolution],
                                                                             'TSemiRegularity': {u'5day': 'annual', u'monthly': None}[temporalResolution],
                                                                             'TCountPerSemiRegularPeriod': {'5day': 73, 'monthly': None}[temporalResolution],
                                                                             'TCornerCoordType': 'min',
                                                                             'IgnoreLeapYear': True,        # CCMP always has a time slice dated March 2, even on a leap year. On non-leap-years, the Feb 26 - Mar 2 slice encompasses 5 days, but on leap years, it encompasses 6 days. We special cased the handling of this in Grid._GetTCoordsList().
                                                                             'PhysicalDimensions': 'tyx',
                                                                             'PhysicalDimensionsFlipped': (False, False, False),
                                                                             'UnscaledDataType': 'int16',
                                                                             'UnscaledNoDataValue': -32767,
                                                                             'ScaledDataType': scaledDataType,
                                                                             'ScaledNoDataValue': {'uwnd': numpy.cast['float32'](0.00152597204*-32767),
                                                                                                   'vwnd': numpy.cast['float32'](0.00152597204*-32767),
                                                                                                   'upstr': numpy.cast['float32'](0.0305194408*-32767),
                                                                                                   'vpstr': numpy.cast['float32'](0.0305194408*-32767),
                                                                                                   'wspd': numpy.cast['float32'](0.00114447903*-32767 + 37.5),
                                                                                                   'nobs': -32767 + 32766}[variableName],
                                                                             'ScalingFunction': {'uwnd': lambda data: numpy.cast['float32'](0.00152597204*data),
                                                                                                 'vwnd': lambda data: numpy.cast['float32'](0.00152597204*data),
                                                                                                 'upstr': lambda data: numpy.cast['float32'](0.0305194408*data),
                                                                                                 'vpstr': lambda data: numpy.cast['float32'](0.0305194408*data),
                                                                                                 'wspd': lambda data: numpy.cast['float32'](0.00114447903*data + 37.5),
                                                                                                 'nobs': lambda data: data + 32766}[variableName],
                                                                             'UnscalingFunction': {'uwnd': lambda data: numpy.cast['uint16'](numpy.round(data/0.00152597204)),
                                                                                                   'vwnd': lambda data: numpy.cast['uint16'](numpy.round(data/0.00152597204)),
                                                                                                   'upstr': lambda data: numpy.cast['uint16'](numpy.round(data/0.0305194408)),
                                                                                                   'vpstr': lambda data: numpy.cast['uint16'](numpy.round(data/0.0305194408)),
                                                                                                   'wspd': lambda data: numpy.cast['uint16'](numpy.round((data - 37.5)/0.00114447903)),
                                                                                                   'nobs': lambda data: numpy.cast['uint16'](data - 32766)}[variableName]})

    def _GetLazyPropertyPhysicalValue(self, name):
        if name == 'CornerCoords':
            return (self.ParentCollection.GetQueryableAttributeValue(u'DateTime'), -78.375, 0.125)
        
        return super(_CCMPWindsL35OPeNDAPGrid, self)._GetLazyPropertyPhysicalValue(name)


class CCMPWindsL35(Grid):
    __doc__ = DynamicDocString()

    def __init__(self, temporalResolution, variableName, productVersion='11', productStream='llk', timeout=60, maxRetryTime=300, cacheDirectory=None):
        # TODO: self.__doc__.Obj.ValidateMethodInvocation()

        displayName = _(u'CCMP L3.5 version %(v1)s.%(v2)s %(stream)s %(temporalResolution)s %(variableDescription)s from NASA PO.DAAC') % \
                      {u'v1': productVersion[0],
                       u'v2': productVersion[1],
                       u'stream': {u'flk': u'first-look', u'llk': u'late-look'}[productStream],
                       u'temporalResolution': temporalResolution,
                       u'variableDescription': {u'uwnd': 'u-component of ocean surface winds',
                                                u'vwnd': 'v-component of ocean surface winds',
                                                u'upstr': 'u-component of ocean surface wind pseudostress',
                                                u'vpstr': 'v-component of ocean surface wind pseudostress',
                                                u'wspd': 'ocean surface wind speed',
                                                u'nobs': 'number of observations',
                                                }[variableName]}

        self._WrappedGrid = TimeSeriesGridStack(_CCMPWindsL35THREDDSCatalog(temporalResolution, displayName, timeout, maxRetryTime, cacheDirectory),
                                                expression = "VariableName = '%s'" % variableName,
                                                reportProgress=False)

        # Unfortunately, 
        self._WrappedGrid.SetLazyPropertyValue('IgnoreLeapYear', True)

    def __getattribute__(self, name):
        if name.startswith('__') or name in ['_WrappedGrid', '_RotateAndClip', 'CreateArcGISRasters', 'CreateClimatologicalArcGISRasters', 'InterpolateAtArcGISPoints']:
            return object.__getattribute__(self, name)
        return object.__getattribute__(object.__getattribute__(self, '_WrappedGrid'), name)

    def __setattr__(self, name, value):
        if name.startswith('__') or name in ['_WrappedGrid', '_RotateAndClip', 'CreateArcGISRasters', 'CreateClimatologicalArcGISRasters', 'InterpolateAtArcGISPoints']:
            object.__setattr__(self, name, value)
        else:
            setattr(object.__getattribute__(self, '_WrappedGrid'), name, value)

    @classmethod
    def _RotateAndClip(cls, grid, rotationOffset, spatialExtent, startDate, endDate):
        if rotationOffset is not None:
            grid = RotatedGlobalGrid(grid, rotationOffset, u'Map units')

        xMin, yMin, xMax, yMax = None, None, None, None
        if spatialExtent is not None:
            from GeoEco.Types import EnvelopeTypeMetadata
            xMin, yMin, xMax, yMax = EnvelopeTypeMetadata.ParseFromArcGISString(spatialExtent)

        if spatialExtent is not None or startDate is not None or endDate is not None:
            if startDate is not None:
                startDate = datetime.datetime(startDate.year, startDate.month, startDate.day, 0, 0, 0)
            if endDate is not None:
                endDate = datetime.datetime(endDate.year, endDate.month, endDate.day, 23, 59, 59)
                
            grid = ClippedGrid(grid, u'Map coordinates', xMin=xMin, xMax=xMax, yMin=yMin, yMax=yMax, tMin=startDate, tMax=endDate)
            
        return grid


###############################################################################
# Metadata: module
###############################################################################

from GeoEco.ArcGIS import ArcGISDependency
from GeoEco.Dependencies import PythonAggregatedModuleDependency
from GeoEco.Datasets.ArcGIS import _UseUnscaledDataDescription, _CalculateStatisticsDescription, _BuildRATDescription, _BuildPyramidsDescription
from GeoEco.Datasets.Virtual import TimeSeriesGridStack, GridSliceCollection, MaskedGrid, ClippedGrid, RotatedGlobalGrid
from GeoEco.Metadata import *
from GeoEco.OceanographicAnalysis.Fronts import CayulaCornillonEdgeDetection, _SIEDDescription, _SIEDReferences, _WindowStatusCodes, _WindowStatusValues
from GeoEco.Types import *

AddModuleMetadata(shortDescription=_(u'THREDDSCatalogs and OPeNDAPURLs for NASA PO.DAAC datasets.'))

###############################################################################
# Metadata: MODISL3SSTTimeSeries class
###############################################################################

_MODISL3SSTTimeSeries_LongDescription = _(

u"""The NASA Jet Propulsion Laboratory (JPL) `Physical Oceanography
Distributed Active Archive Center <http://podaac.jpl.nasa.gov/>`_
(PO.DAAC) publishes collections of sea surface temperature (SST)
images gathered by the
`Moderate Resolution Imaging Spectroradiometer <http://podaac.jpl.nasa.gov/SeaSurfaceTemperature/MODIS>`_
(MODIS) carried by the Terra and Aqua satellites. This tool accesses
the PO.DAAC SST datasets that begin with the following names:

* MODIS Terra Level 3 Thermal IR
* MODIS Terra Level 3 Mid-IR
* MODIS Aqua Level 3 Thermal IR
* MODIS Aqua Level 3 Mid-IR
""")

_MODISReferences = _(
u"""**References**

The `PO.DAAC MODIS Level 3 User Guide <ftp://podaac-ftp.jpl.nasa.gov/allData/modis/L3/docs/UserGuide_MODIS_L3_v2014.0.pdf>`_
provides more information about the MODIS SST datasets.

Please see
`these instructions <http://podaac.jpl.nasa.gov/CitingPODAAC>`_
on how to acknowledge your use of PO.DAAC products.""")

AddClassMetadata(MODISL3SSTTimeSeries,
    shortDescription=_(u'Time series of MODIS SST images published by NASA JPL PO.DAAC.'),
    longDescription=_MODISL3SSTTimeSeries_LongDescription + _(
u"""
PO.DAAC presents each collection as a time series of 2D HDF files that
may be accessed through a variety of protocols, including `OPeNDAP <http://opendap.org/>`_.
Given a satellite name, temporal resolution, spatial resolution, and
desired geophysical parameter, this class efficiently accesses the
specified time series of images using OPeNDAP, representing them as a
single 3D grid with dimensions x, y, and time.

""") + _MODISReferences)

# Constructor

AddMethodMetadata(MODISL3SSTTimeSeries.__init__,
    shortDescription=_(u'Constructs a new MODISL3SSTTimeSeries instance.'),
    isExposedToPythonCallers=True,
    dependencies=[PythonAggregatedModuleDependency('numpy')])

AddArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'self',
    typeMetadata=ClassInstanceTypeMetadata(cls=MODISL3SSTTimeSeries),
    description=_(u'MODISL3SSTTimeSeries instance.'))

AddArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'satellite',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Aqua', u'Terra'], makeLowercase=True),
    description=_(
u"""MODIS satellite to use, one of:

* Aqua - the Aqua datasets start in July 2002.

* Terra - the Terra datasets start in February 2000.
"""),
    arcGISDisplayName=_(u'Satellite'))

AddArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'temporalResolution',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Daily', u'8day', u'Monthly', u'Annual'], makeLowercase=True),
    description=_(
u"""Temporal resolution to use, one of:

* Daily - daily images. There are 365 during normal years and 366
  during leap years.

* 8day - 8-day images. There are 46 per year. The first image of the
  year starts on January 1. The duration of the last image of the year
  is five days during normal years and six days during leap years.

* Monthly - monthly images.

* Annual - annual images.

Although NASA may publish MODIS SST images at other temporal
resolutions, they are not supported at this time. If you need one of those
products, please contact the author of this tool to see if support may
be added.

The satellites experience occasional transient failures that prevent
data from being collected. NASA opted not to produce any images for
these periods. These missing images are represented as time slices
filled with the NoData value."""),
    arcGISDisplayName=_(u'Temporal resolution'))

AddArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'spatialResolution',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'4km', u'9km'], makeLowercase=True),
    description=_(
u"""Spatial resolution to use, one of:

* 4km - the grid has a cell size of 1/24 geographic degree, or about
  4.64 km at the equator, with 8640 columns and 4320 rows. 

* 9km - the grid has a cell size of 1/12 geographic degree, or about
  9.28 km at the equator, with 4320 columns and 2160 rows.
"""),
    arcGISDisplayName=_(u'Spatial resolution'))

AddArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'geophysicalParameter',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'SST', u'NSST', u'SST4'], makeLowercase=True),
    description=_(
u"""Geophysical parameter to use, one of:

* SST - daytime SST derived from the 11 and 12 micrometer thermal IR
  infrared (IR) bands (MODIS channels 31 and 32)

* NSST - nighttime SST derived from the 11 and 12 micrometer thermal
  IR infrared (IR) bands (MODIS channels 31 and 32)

* SST4 -nighttime SST derived from the 3.8-4.1 micrometer mid-IR
  infrared (IR) bands (MODIS channels 20, 22, and 23)

The geophysical parameter determines the time of day that the data
apply to. During the daytime, the satellites pass over the equator at
slightly different times, 10:30 for Terra and 13:30 for Aqua, but for
simplicity it is assumed that they both pass over at 12:00. Therefore,
a daily image of the SST parameter (daytime temperature) is assumed to
apply from 00:00 to 23:59 on the date of the image. The nighttime
passes occur 12 hours earlier. Thus a daily image of the NSST or SST4
parameter (nighttime temperature) is assumed to apply from 12:00 on
the previous day to 11:59 of the date of the image. This scheme
reflects what is described by the Start Time and End Time global
attributes of the MODIS L3 HDF files."""),
    arcGISDisplayName=_(u'Geophysical parameter'))

AddArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'variableName',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'sst', u'quality'], makeLowercase=True),
    description=_(
u"""Dataset variable to use, one of:

* sst - the value of the geophysical parameter (SST). The data type
  will be 32-bit floating point, or 16-bit unsigned integers if the
  original unscaled values are requested.

* quality - the pixel quality level. The data type will be integer
  (see below). According to the
  `PO.DAAC MODIS Level 3 User Guide <ftp://podaac-ftp.jpl.nasa.gov/allData/modis/L3/docs/UserGuide_MODIS_L3_v2014.0.pdf>`_
  the quality level can theoretically range from 0 (best) to 4
  (worst). In practice, NASA automatically masks all pixels in both
  the SST image and the quality level image where the quality is 3 or
  4, thus you are only likely to encounter values 0, 1, or 2 in
  quality level images produced by this tool. In general, the quality
  level 0 indicates a good SST value; 1 indicates a questionable value
  due to the pixel being too near to the edge of the sensor's field of
  view or the 3x3 neighborhood around the pixel showing too high a
  range of values; 2 indicates a value that is probably bad due to
  clouds or other problems; 3 and 4 indicate values that are
  definitely bad due to clouds or land.
"""),
    arcGISDisplayName=_(u'Dataset variable'))

AddArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'qualityLevel',
    typeMetadata=IntegerTypeMetadata(canBeNone=True, allowedValues=[0, 1, 2, 3, 4]),
    description=_(
u"""Worst acceptable pixel pixel quality level. Pixels with worse
quality than this will be masked.

According to the
`PO.DAAC MODIS Level 3 User Guide <ftp://podaac-ftp.jpl.nasa.gov/allData/modis/L3/docs/UserGuide_MODIS_L3_v2014.0.pdf>`_
the quality level can theoretically range from 0 (best) to 4 (worst).
If you pick 1 as the worst acceptable level, for example, then pixels
with quality levels 2, 3, and 4 will be masked. In practice, NASA
automatically masks all pixels where the quality is 3 or 4, thus
selecting 3 or 4 as the worst acceptable level will have no effect, as
those pixels will have already been masked by NASA before NASA makes
the data available for download, and there is no way to unmask them.

In general, the quality level 0 indicates a good SST value; 1
indicates a questionable value due to the pixel being too near to the
edge of the sensor's field of view or the 3x3 neighborhood around the
pixel showing too high a range of values; 2 indicates a value that is
probably bad due to clouds or other problems; 3 and 4 indicate values
that are definitely bad due to clouds or land."""),
    arcGISDisplayName=_(u'Pixel quality level'),
    arcGISCategory=_(u'Masking options'))

CopyArgumentMetadata(THREDDSCatalog.__init__, u'timeout', MODISL3SSTTimeSeries.__init__, u'timeout')
CopyArgumentMetadata(THREDDSCatalog.__init__, u'maxRetryTime', MODISL3SSTTimeSeries.__init__, u'maxRetryTime')
CopyArgumentMetadata(THREDDSCatalog.__init__, u'cacheDirectory', MODISL3SSTTimeSeries.__init__, u'cacheDirectory')

AddResultMetadata(MODISL3SSTTimeSeries.__init__, u'grid',
    typeMetadata=ClassInstanceTypeMetadata(cls=MODISL3SSTTimeSeries),
    description=_(u'MODISL3SSTTimeSeries instance.'))

# Public method: MODISL3SSTTimeSeries.CreateArcGISRasters

AddMethodMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters,
    shortDescription=_(u'Creates rasters for MODIS Level 3 SST images published by NASA JPL PO.DAAC.'),
    longDescription=_MODISL3SSTTimeSeries_LongDescription + _(
u"""
Given a satellite name, temporal resolution, spatial resolution, and
desired geophysical parameter, this tool efficiently downloads the
specified images using the `OPeNDAP <http://opendap.org/>`_ protocol
and creates corresponding rasters.

""") + _MODISReferences,
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Create Rasters for PO.DAAC MODIS L3 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\MODIS Global Level 3 Mapped SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cls',
    typeMetadata=ClassOrClassInstanceTypeMetadata(cls=MODISL3SSTTimeSeries),
    description=_(u'MODISL3SSTTimeSeries class or instance.'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'satellite', MODISL3SSTTimeSeries.CreateArcGISRasters, u'satellite')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'temporalResolution', MODISL3SSTTimeSeries.CreateArcGISRasters, u'temporalResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'spatialResolution', MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'geophysicalParameter', MODISL3SSTTimeSeries.CreateArcGISRasters, u'geophysicalParameter')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'variableName', MODISL3SSTTimeSeries.CreateArcGISRasters, u'variableName')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'outputWorkspace',
    typeMetadata=ArcGISWorkspaceTypeMetadata(createParentDirectories=True),
    description=_(
u"""Directory or geodatabase to receive the rasters.

Unless you have a specific reason to store the rasters in a
geodatabase, we recommend you store them in a directory because it
will be much faster and allows the rasters to be organized in a tree.
If you do store the rasters in a geodatabase, you must change the
Raster Name Expressions parameter; see below for more
information."""),
    arcGISDisplayName=_(u'Output workspace'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'mode',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Add', u'Replace'], makeLowercase=True),
    description=_(
u"""Overwrite mode, one of:

* Add - create rasters that do not exist and skip those that already
  exist. This is the default.

* Replace - create rasters that do not exist and overwrite those that
  already exist.

'Add' should be appropriate for nearly all situations. One situation
in which 'Replace' is appropriate is when the data publisher
reprocesses the entire dataset with improved algorithms and releases
updated images. In that case, you may wish to recreate all of your
rasters using the updated data.

The ArcGIS Overwrite Outputs geoprocessing setting has no effect on
this tool. If 'Replace' is selected the rasters will be overwritten,
regardless of the ArcGIS Overwrite Outputs setting."""),
    arcGISDisplayName=_(u'Overwrite mode'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure with names that
imitate those used by NASA. When storing rasters in a directory, the
final expression specifies the file name of the raster and any
preceding expressions specify subdirectories. The extension of the
final expression determines the output raster format: .asc for ArcInfo
ASCII Grid, .bmp for BMP, .gif for GIF, .img for an ERDAS IMAGINE
file, .jpg for JPEG, .jp2 for JPEG 2000, .png for PNG, .tif for
GeoTIFF, or no extension for ArcInfo Binary Grid. The default
expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(Satellite)s - name of the satellite, either "aqua" or "terra".

* %(SatelliteCode)s - abbreviation for the satellite, either "A" or
  "T", used in NASA's file naming scheme.

* %(TemporalResolution)s - temporal resolution, either "daily",
  "8day", "monthly", or "annual".

* %(TemporalResolutionCode)s - abbreviation for the temporal
  resolution, either "DAY", "8D", "MO", or "YR", used in NASA's file
  naming scheme.

* %(SpatialResolution)s - spatial resolution, either "4km" or "9km".

* %(SpatialResolutionCode)s - abbreviation for the spatial
  resolution, either "4" or "9", used in NASA's file naming scheme.
  
* %(GeophysicalParameter)s - geophysical parameter represented in the
  output raster, either "sst", "nsst", or "sst4".

* %(Algorithm)s - algorithm used to estimate the geophysical
  parameter from the raw sensor values, either "sst" for the sst or
  nsst geophysical parameter, or "sst4" for the sst4 geophysical
  parameter. The algorithm name is used in NASA's file naming scheme.

* %(Wavelength)s - an alternative abbreviation for the algorithm used
  to estimate the geophysical parameter, either "11um" for the sst or
  nsst geophysical parameter, or "4um" for the sst4 geophysical
  parameter. The abbreviation refers to the wavelengths of the MODIS
  bands used in the algorithm, where 11um refers to the "thermal IR"
  bands and 4um refers to the "mid-IR" bands. The abbreviation is used
  in NASA's file naming scheme.

* %(VariableName)s - dataset variable represented in the output
  raster, either the SST data or the pixel quality level.

* %%Y - four-digit year of the raster.

* %%m - two-digit month of the first day of the raster.

* %%d - two-digit day of the month of the first day of the raster.

* %%j - three-digit day of the year of the first day of the raster.

* %(EndDate)s - date of the last day of the raster in the format
  "YYYYjjj" where YYYY is the four-digit year and jjj is the
  three-digit day of the year. This date is used in NASA's file naming
  scheme for temporal resolutions other than "daily".
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rasterCatalog',
    typeMetadata=ArcGISRasterCatalogTypeMetadata(canBeNone=True, mustBeDifferentThanArguments=[u'outputWorkspace'], createParentDirectories=True),
    description=_(
u"""Raster catalog to create.

This parameter requires ArcGIS 9.3 or later.

If this parameter is specified, after the tool finishes creating
rasters, it will create an unmanaged raster catalog and import all of
the rasters in the output workspace into it. The catalog will have
fields for all of the codes specified in the Raster Name Expressions
as well as fields for the start date, center date, and end date of
each raster. You can then use the catalog to create animations using
the ArcGIS 10 Time Slider.

WARNING: The raster catalog will be deleted and re-created each time
the tool is executed. Beware of this if you plan to add your own
fields to the catalog after it has been created."""),
    direction=u'Output',
    arcGISDisplayName=_(u'Output raster catalog'),
    dependencies=[ArcGISDependency(9, 3)])

CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'qualityLevel', MODISL3SSTTimeSeries.CreateArcGISRasters, u'qualityLevel')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rotationOffset',
    typeMetadata=FloatTypeMetadata(canBeNone=True),
    description=_(
u"""Degrees to rotate the output rasters about the polar axis. If not
provided, the output rasters will be centered on the Prime Meridian
and have x coordinates ranging from -180 to 180.

Use this parameter to shift the center longitude to a different
location. Positive values shift it to the east, negative values to the
west. For example, to center the output rasters on the Pacific ocean
and use x coordinates ranging from 0 to 360 rather than -180 to 180,
provide 180 for this parameter.

The rasters can only be rotated in whole grid cells. The value you
provide will be rounded off to the closest cell."""),
    arcGISDisplayName=_(u'Rotate raster by'),
    arcGISCategory=_(u'Spatiotemporal extent'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialExtent',
    typeMetadata=EnvelopeTypeMetadata(canBeNone=True),
    description=_(
u"""Spatial extent of the output rasters, in degrees. If not provided,
the spatial extent will not be clipped.

This parameter is applied after the rotation parameter and uses
coordinates that result after rotation. For example, if the rotation
parameter was 180, the resulting rasters will have x coordinates
ranging from 0 to 360. The spatial extent should be expressed in those
coordinates.

The rasters can only be clipped in whole grid cells. The values you
provide will be rounded off to the closest cell."""),
    arcGISDisplayName=_(u'Spatial extent'),
    arcGISCategory=_(u'Spatiotemporal extent'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'startDate',
    typeMetadata=DateTimeTypeMetadata(canBeNone=True),
    description=_(
u"""Start date for the rasters to create. Rasters will be created for
images that occur on or after the start date and on or before the end
date. If the start date is not provided, the date of the oldest image
will be used.

The time component of the start date is ignored."""),
    arcGISDisplayName=_(u'Start date'),
    arcGISCategory=_(u'Spatiotemporal extent'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'endDate',
    typeMetadata=DateTimeTypeMetadata(canBeNone=True),
    description=_(
u"""End date for the rasters to create. Rasters will be created for
images that occur on or after the start date and on or before the end
date. If the end date is not provided, the date of the most recent
image will be used.

The time component of the end date is ignored."""),
    arcGISDisplayName=_(u'End date'),
    arcGISCategory=_(u'Spatiotemporal extent'))

CopyArgumentMetadata(THREDDSCatalog.__init__, u'timeout', MODISL3SSTTimeSeries.CreateArcGISRasters, u'timeout')
CopyArgumentMetadata(THREDDSCatalog.__init__, u'maxRetryTime', MODISL3SSTTimeSeries.CreateArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(THREDDSCatalog.__init__, u'cacheDirectory', MODISL3SSTTimeSeries.CreateArcGISRasters, u'cacheDirectory')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'useUnscaledData',
    typeMetadata=BooleanTypeMetadata(),
    description=_UseUnscaledDataDescription,
    arcGISDisplayName=_(u'Use unscaled data'),
    arcGISCategory=_(u'Additional raster processing options'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'calculateStatistics',
    typeMetadata=BooleanTypeMetadata(),
    description=_CalculateStatisticsDescription,
    arcGISDisplayName=_(u'Calculate statistics'),
    arcGISCategory=_(u'Additional raster processing options'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildRAT',
    typeMetadata=BooleanTypeMetadata(),
    description=_BuildRATDescription,
    arcGISDisplayName=_(u'Build raster attribute tables'),
    arcGISCategory=_(u'Additional raster processing options'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildPyramids',
    typeMetadata=BooleanTypeMetadata(),
    description=_BuildPyramidsDescription,
    arcGISDisplayName=_(u'Build pyramids'),
    arcGISCategory=_(u'Additional raster processing options'))

AddResultMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'updatedOutputWorkspace',
    typeMetadata=ArcGISWorkspaceTypeMetadata(),
    description=_(u'Updated output workspace.'),
    arcGISDisplayName=_(u'Updated output workspace'),
    arcGISParameterDependencies=[u'outputWorkspace'])

# Public method: MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters

AddMethodMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters,
    shortDescription=_(u'Creates climatological rasters from MODIS Level 3 SST images published by NASA JPL PO.DAAC.'),
    longDescription=_MODISL3SSTTimeSeries_LongDescription + _(
u"""
This tool produces rasters showing the climatological average value
(or other statistic) of a time series of MODIS SST images. Given a
specification of the desired MODIS SST time series, a statistic, and a
climatological bin definition, this tool efficiently downloads the
images using the `OPeNDAP <http://opendap.org/>`_ protocol, classifies
them into bins, and produces a single raster for each bin. Each cell
of the raster is produced by calculating the statistic on the values
of that cell extracted from all of the rasters in the bin.

""") + _MODISReferences,
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Create Climatological Rasters for PO.DAAC MODIS L3 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\MODIS Global Level 3 Mapped SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cls', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'cls')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'satellite', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'satellite')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'temporalResolution', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'temporalResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialResolution', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'spatialResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'geophysicalParameter', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'geophysicalParameter')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'variableName', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'variableName')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'statistic',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Count', u'Maximum', u'Mean', u'Minimum', u'Range', u'Standard Deviation', u'Sum'], makeLowercase=True),
    description=_(
u"""Statistic to calculate for each cell, one of:

* Count - number of images in which the cell had data.

* Maximum - maximum value for the cell.

* Mean - mean value for the cell, calculated as the sum divided by the
  count.

* Minimum - minimum value for the cell.

* Range - range for the cell, calculated as the maximum minus the
  minimum.

* Standard Deviation - sample standard deviation for the cell
  (i.e. the standard deviation estimated using Bessel's correction).
  In order to calculate this, there must be at least two images with
  data for the cell.

* Sum - the sum for the cell.
"""),
    arcGISDisplayName=_(u'Statistic'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'binType',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Daily', u'Monthly', u'Cumulative', u'ENSO Daily', u'ENSO Monthly', u'ENSO Cumulative'], makeLowercase=True),
    description=_(
u"""Climatology bins to use, one of:

* Daily - daily bins. Images will be classified into bins according to
  their days of the year. The number of days in each bin is determined
  by the Climatology Bin Duration parameter (which defaults to 1). The
  number of bins is calculated by dividing 365 by the bin duration. If
  there is no remainder, then that number of bins will be created;
  images for the 366th day of leap years will be counted in the bin
  that includes day 365. For example, if the bin duration is 5, 73
  bins will be created. The first will be for days 1-5, the second
  will be for days 5-10, and so on; the 73rd bin will be for days
  361-365 during normal years and 361-366 during leap years. If
  dividing 365 by the bin duration does yield a remainder, then one
  additional bin will be created to hold the remaining days. For
  example, if the bin duration is 8, 46 bins will be created. The
  first will be for days 1-8, the second for days 9-16, and so on; the
  46th will be for days 361-365 during normal years and 361-366 during
  leap years.

* Monthly - monthly bins. Images will be classified into bins according to
  their months of the year. The number of months in each bin is
  determined by the Climatology Bin Duration parameter (which defaults
  to 1). The number of bins is calculated by dividing 12 by the bin
  duration. If there is no remainder, then that number of bins will be
  created. For example, if the bin duration is 3, there will be four
  bins: January-March, April-June, July-September, and
  October-December. If there is a remainder, then one additional bin
  will be created. For example, if the bin duration is 5, 3 bins will
  be created: January-May, June-October, November-December.

* Cumulative - one bin. A single climatology raster will be calculated
  from the entire dataset. The Bin Duration parameter is ignored.

* ENSO Daily, ENSO Monthly, ENSO Cumulative - the same as above,
  except each of the bins above will be split into three, based on the
  phase of the `El Nino Southern Oscillation (ENSO) <http://en.wikipedia.org/wiki/ENSO>`_,
  as determined by the Oceanic Nino Index (ONI) calculated by the
  `NOAA NCEP Climate Prediction Center <http://www.cpc.ncep.noaa.gov/products/analysis_monitoring/ensostuff/ensoyears.shtml>`_.
  The ONI classifies each month into one of three phases: neutral, El
  Nino, or La Nina. This tool first classifies input images according
  to their dates into ENSO phases (it downloads ONI data from the
  `NOAA Earth System Research Laboratory <http://www.esrl.noaa.gov/psd/data/climateindices/list/>`_),
  then produces a climatology bin for each phase. For example, if you
  request ENSO Cumulative bins, three bins will be produced: one for
  all images occurring in neutral months, one for all in El Nino
  months, and one for all in La Nina months. If you request ENSO
  Monthly bins, 36 bins will be produced: one for each combination of
  the 12 months and the three ENSO phases.

For Daily and Monthly, to adjust when the bins start (e.g. to center a
4-bin seasonal climatology on solstices and equinoxes), use the Start
Climatology At This Day Of The Year parameter."""),
    arcGISDisplayName=_(u'Climatology bin type'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'outputWorkspace', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'outputWorkspace')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'mode',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Add', u'Replace'], makeLowercase=True),
    description=_(
u"""Overwrite mode, one of:

* Add - create rasters that do not exist and skip those that already
  exist. This is the default.

* Replace - create rasters that do not exist and overwrite those that
  already exist. Choose this option when you want to regenerate the
  climatologies using the latest images.

The ArcGIS Overwrite Outputs geoprocessing setting has no effect on
this tool. If 'Replace' is selected the rasters will be overwritten,
regardless of the ArcGIS Overwrite Outputs setting."""),
    arcGISDisplayName=_(u'Overwrite mode'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure. When storing
rasters in a directory, the final expression specifies the file name
of the raster and any preceding expressions specify subdirectories.
The extension of the final expression determines the output raster
format: .asc for ArcInfo ASCII Grid, .bmp for BMP, .gif for GIF, .img
for an ERDAS IMAGINE file, .jpg for JPEG, .jp2 for JPEG 2000, .png for
PNG, .tif for GeoTIFF, or no extension for ArcInfo Binary Grid. The
default expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(Satellite)s - name of the satellite, either "aqua" or "terra".

* %(SatelliteCode)s - abbreviation for the satellite, either "A" or
  "T", used in NASA's file naming scheme.

* %(TemporalResolution)s - temporal resolution, either "daily",
  "8day", "monthly", or "annual".

* %(TemporalResolutionCode)s - abbreviation for the temporal
  resolution, either "DAY", "8D", "MO", or "YR", used in NASA's file
  naming scheme.

* %(SpatialResolution)s - spatial resolution, either "4km" or "9km".

* %(SpatialResolutionCode)s - abbreviation for the spatial
  resolution, either "4" or "9", used in NASA's file naming scheme.
  
* %(GeophysicalParameter)s - geophysical parameter represented in the
  output raster, either "sst", "nsst", or "sst4".

* %(Algorithm)s - algorithm used to estimate the geophysical
  parameter from the raw sensor values, either "sst" for the sst or
  nsst geophysical parameter, or "sst4" for the sst4 geophysical
  parameter. The algorithm name is used in NASA's file naming scheme.

* %(Wavelength)s - an alternative abbreviation for the algorithm used
  to estimate the geophysical parameter, either "11um" for the sst or
  nsst geophysical parameter, or "4um" for the sst4 geophysical
  parameter. The abbreviation refers to the wavelengths of the MODIS
  bands used in the algorithm, where 11um refers to the "thermal IR"
  bands and 4um refers to the "mid-IR" bands. The abbreviation is used
  in NASA's file naming scheme.

* %(VariableName)s - dataset variable represented in the output
  raster, either the SST data or pixel quality level.

* %(ClimatologyBinType)s - type of the climatology bin, either "Daily"
  if 1-day bins, "Xday" if multi-day bins (X is replaced by the
  duration), "Monthly" if 1-month bins, "Xmonth" if multi-month bins,
  or "Cumulative". If an ENSO bin type is used, "ENSO\\_" will be
  prepended to those strings (e.g. "ENSO_Daily", "ENSO_Monthly").

* %(ClimatologyBinName)s - name of the climatology bin corresponding
  represented by the output raster, either "dayXXX" for 1-day bins
  (XXX is replaced by the day of the year), "daysXXXtoYYY" for
  multi-day bins (XXX is replaced by the first day of the bin, YYY is
  replaced by the last day), "monthXX" for 1-month bins (XX is
  replaced by the month), "monthXXtoYY" (XX is replaced by the first
  month of the bin, YY by the last month), or "cumulative". If an ENSO
  bin type is used, "neutral\\_", "ElNino\\_", and "LaNina\\_" will be
  prepended to those strings for each of the three ENSO phased rasters
  (e.g. "neutral_cumulative", "ElNino_cumulative", and
  "LaNina_cumulative" when "ENSO Cumulative" bins are requested).

* %(Statistic)s - statistic that was calculated, in lowercase and with
  spaces replaced by underscores; one of: "count", "maximum", "mean",
  "minimum", "range", "standard_deviation", "Sum".

If the Bin Type is "Daily", the following additional codes are
available:

* %(FirstDay)i - first day of the year of the climatology bin
  represented by the output raster.

* %(LastDay)i - last day of the year of the climatology bin
  represented by the output raster. For 1-day climatologies, this will
  be the same as %(FirstDay)i.

If the Bin Type is "Monthly", the following additional codes are
available:

* %(FirstMonth)i - first month of the climatology bin represented by
  the output raster.

* %(DayOfFirstMonth)i - first day of the first month of the
  climatology bin represented by the output raster.

* %(LastMonth)i - last month of the climatology bin represented by
  the output raster.

* %(DayOfLastMonth)i - last day of the last month of the climatology
  bin represented by the output raster.

Note that the additional codes are integers and may be formatted using
"printf"-style formatting codes. For example, to format the FirstDay
as a three-digit number with leading zeros::

    %(FirstDay)03i
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'qualityLevel', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'qualityLevel')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'binDuration',
    typeMetadata=IntegerTypeMetadata(minValue=1),
    description=_(
u"""Duration of each bin, in days or months, when the Bin Type is
Daily or Monthly, respectively. The default is 1. See the Bin Type
parameter for more information."""),
    arcGISDisplayName=_(u'Climatology bin duration'),
    arcGISCategory=_(u'Climatology options'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'startDayOfYear',
    typeMetadata=IntegerTypeMetadata(minValue=1, maxValue=365),
    description=_(
u"""Use this parameter to create bin defintions that deviate from the
traditional calendar. The interpretation of this parameter depends on
the Bin Type:

* Daily - this parameter defines the day of the year of the first
  climatology bin. For example, if this parameter is 100 and the Bin
  Duration is 10, the first bin will be numbered 100-109. The bin
  spanning the end of the year will be numbered 360-004. The last bin
  will be numbered 095-099. To define a four-bin climatology with bins
  that are centered approximately on the equinoxes and solstices
  (i.e., a seasonal climatology), set the Bin Duration to 91 and the
  start day to 36 (February 5). This will produce bins with dates
  036-126, 127-217, 218-308, and 309-035.

* Monthly - this parameter defines the day of the year of the first
  climatology bin, and the day of the month of that bin will be used
  as the first day of the month of all of the bins. For example, if
  this parameter is 46, which is February 15, and the Bin Duration is
  1, then the bins will be February 15 - March 14, March 15 - April
  14, April 15 - May 14, and so on. Calculations involving this
  parameter always assume a 365 day year (a non-leap year). To define
  a four-bin climatology using the months traditionally associated
  with spring, summer, fall, and winter in many northern hemisphere
  cultures, set the Bin Duration to 3 and the start day to 60 (March
  1). This will produce bins with months 03-05, 06-08, 09-11, and
  12-02.

* Cumulative - this parameter is ignored.
"""),
    arcGISDisplayName=_(u'Start climatology at this day of the year'),
    arcGISCategory=_(u'Climatology options'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rotationOffset', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'rotationOffset')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialExtent', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'spatialExtent')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'startDate', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'startDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'endDate', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'endDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'timeout', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'timeout')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'maxRetryTime', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cacheDirectory', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'cacheDirectory')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'calculateStatistics', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'calculateStatistics')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildPyramids', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'buildPyramids')

CopyResultMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'updatedOutputWorkspace', MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'updatedOutputWorkspace')

# Public method: MODISL3SSTTimeSeries.InterpolateAtArcGISPoints

AddMethodMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints,
    shortDescription=_(u'Interpolates PO.DAAC MODIS Level 3 SST values at points.'),
    longDescription=_MODISL3SSTTimeSeries_LongDescription + _(
u"""
Given a satellite name, temporal resolution, spatial resolution, and
desired geophysical parameter, this tool interpolates the value of
that parameter at the given points. This tool performs the same basic
operation as the ArcGIS Spatial Analyst's Extract Values to Points
tool, but it reads the MODIS data directly from NASA's servers using
the `OPeNDAP <http://opendap.org/>`_ protocol, rather than reading
rasters stored on your machine.

""") + _MODISReferences,
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Interpolate PO.DAAC MODIS L3 SST at Points'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\MODIS Global Level 3 Mapped SST'),
    dependencies=[ArcGISDependency(9, 1, requiresCOMInstantiation=True), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cls', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'cls')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'satellite', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'satellite')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'temporalResolution', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'temporalResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'spatialResolution', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'spatialResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'geophysicalParameter', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'geophysicalParameter')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'variableName', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'variableName')

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'points',
    typeMetadata=ArcGISFeatureLayerTypeMetadata(mustExist=True, allowedShapeTypes=[u'Point']),
    description=_(
u"""Points at which values should be interpolated.

The SST data use the WGS 1984 geographic coordinate system. It is
recommended but not required that the points use the same coordinate
system. If they do not, this tool will attempt to project the points
to the WGS 1984 coordinate system prior to doing the interpolation.
This may fail if a datum transformation is required, in which case you
will have to manually project the points to the WGS 1984 coordinate
system before using this tool."""),
    arcGISDisplayName=_(u'Point features'))

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'valueField',
    typeMetadata=ArcGISFieldTypeMetadata(mustExist=True, allowedFieldTypes=[u'short', u'long', u'float', u'double']),
    description=_(
u"""Field of the points to receive the interpolated values.

The field must have a floating-point or integer data type. If the
field cannot represent the interpolated value at full precision, the
closest approximation will be stored and a warning will be issued.
This will happen, for example, when you interpolate SST values into an
integer field."""),
    arcGISDisplayName=_(u'Field to receive the interpolated values'),
    arcGISParameterDependencies=[u'points'])

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'tField',
    typeMetadata=ArcGISFieldTypeMetadata(mustExist=True, allowedFieldTypes=[u'date']),
    description=_(
u"""Field of the points that specifies the date and time of the point.

The field must have a date or datetime data type. If the field can
only represent dates with no time component, the time will assumed to
be 00:00:00."""),
    arcGISDisplayName=_(u'Date field'),
    arcGISParameterDependencies=[u'points'])

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'method',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Nearest', u'Linear'], makeLowercase=True),
    description=_(
u"""Interpolation method to use, one of:

* Nearest - nearest neighbor interpolation. The interpolated value
  will simply be the value of the cell that contains the point. This
  is the default.

* Linear - linear interpolation (also known as trilinear
  interpolation). This method is suitable for the SST data, which are
  continuous, but is not appropriate for the pixel quality levels,
  which are categorical (use nearest neighbor instead). This method
  averages the values of the eight nearest cells in the x, y, and time
  dimensions, weighting the contribution of each cell by the area of
  it that would be covered by a hypothetical cell centered on the
  point being interpolated. If the cell containing the point contains
  NoData, the result is NoData. If any of the other seven cells
  contain NoData, they are omitted from the average, and the result is
  based on the weighted average of the cells that do contain data.
  This is the same algorithm implemented by the ArcGIS Spatial
  Analyst's Extract Values to Points tool.
"""),
    arcGISDisplayName=_(u'Interpolation method'))

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'where',
    typeMetadata=SQLWhereClauseTypeMetadata(canBeNone=True),
    description=_(
u"""SQL WHERE clause expression that specifies the subset of points to
use. If this parameter is not provided, all of the points will be
used.

The exact syntax of this expression depends on the type of feature
class you're using. ESRI recommends you reference fields using the
following syntax:

* For shapefiles, ArcInfo coverages, or feature classes stored in file
  geodatabases, ArcSDE geodatabases, or ArcIMS, enclose field names in
  double quotes: "MY_FIELD"

* For feature classes stored in personal geodatabases, enclose field
  names in square brackets: [MY_FIELD].
"""),
    arcGISDisplayName=_(u'Where clause'),
    arcGISCategory=_(u'Interpolation options'),
    arcGISParameterDependencies=[u'points'])

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'noDataValue',
    typeMetadata=FloatTypeMetadata(canBeNone=True),
    description=_(
u"""Value to use when the interpolated value is NoData.

If a value is not provided for this parameter, a database NULL value
will be stored in the field when the interpolated value is NoData. If
the field cannot store NULL values, as is the case with shapefiles,
the value -9999 will be used."""),
    arcGISDisplayName=_(u'Value to use when the interpolated value is NoData'),
    arcGISCategory=_(u'Interpolation options'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'qualityLevel', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'qualityLevel')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'timeout', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'timeout')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'maxRetryTime', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'maxRetryTime')
CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'cacheDirectory', MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'cacheDirectory')

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'orderByFields',
    typeMetadata=ListTypeMetadata(elementType=ArcGISFieldTypeMetadata(mustExist=True), minLength=1, canBeNone=True),
    description=_(
u"""Fields for defining the order in which the points are processed.

The points may be processed faster if they are ordered
spatiotemporally, such that points that are close in space and time
are processed sequentially. Ordering the points this way increases the
probability that the value of a given point can be interpolated from
data that is cached in memory, rather than from data that must be read
from the disk or network, which is much slower. Choose fields that
faciliate this. For example, if your points represent the locations of
animals tracked by satellite telemetry, order the processing first by
the animal ID and then by the transmission date or number.

If you omit this parameter, the Date Field will be used automatically.

This parameter requires ArcGIS 9.2 or later."""),
    arcGISDisplayName=_(u'Order by fields'),
    arcGISCategory=_(u'Performance tuning options'),
    arcGISParameterDependencies=[u'points'],
    dependencies=[ArcGISDependency(9, 2)])

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'numBlocksToCacheInMemory',
    typeMetadata=IntegerTypeMetadata(minValue=0, canBeNone=True),
    description=_(
u"""Maximum number of blocks of SST data to cache in memory.

To minimize the number of times that the disk or network must be
accessed, this tool employs a simple caching strategy, in addition to
disk caching described by the Cache Directory parameter. When it
processes the first point, it reads a square block of cells centered
on that point and caches it in memory. When it processes the second
and subsequent points, it first checks whether the cells needed for
that point are contained by the block cached in memory. If so, it
processes that point using the in-memory block, rather than reading
from disk or the network again. If not, it reads another square block
centered on that point and adds it to the cache.

The tool processes the remaining points, adding additional blocks to
the cache, as needed. To prevent the cache from exhausing all memory,
it is only permitted to grow to the size specified by this parameter.
When the cache is full but a new block is needed, the oldest block is
discarded to make room for the newest block.

The maximum size of the cache in bytes may be calculated by
multiplying this parameter by the block size parameters and the data
type of the data being accessed (e.g. 4 bytes for the SST data). For
example, if this parameter is 128 and the blocks are x=32 by y=32 by
t=2 and you're interpolating the SST data which are 4 bytes, the
maximum size of the cache is 1048576 bytes (1 MB).

If this parameter is 0, no blocks will be cached in memory."""),
    arcGISDisplayName=_(u'Number of blocks of data to cache in memory'),
    arcGISCategory=_(u'Performance tuning options'))

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'xBlockSize',
    typeMetadata=IntegerTypeMetadata(minValue=0, canBeNone=True),
    description=_(
u"""Size of the blocks of SST data to cache in memory, in the x
direction (longitude). The size is given as the number of cells.

If this parameter is 0, no blocks will be cached in memory."""),
    arcGISDisplayName=_(u'In-memory cache block size, in X direction'),
    arcGISCategory=_(u'Performance tuning options'))

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'yBlockSize',
    typeMetadata=IntegerTypeMetadata(minValue=0, canBeNone=True),
    description=_(
u"""Size of the blocks of SST data to cache in memory, in the y
direction (latitude). The size is given as the number of cells.

If this parameter is 0, no blocks will be cached in memory."""),
    arcGISDisplayName=_(u'In-memory cache block size, in Y direction'),
    arcGISCategory=_(u'Performance tuning options'))

AddArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'tBlockSize',
    typeMetadata=IntegerTypeMetadata(minValue=0, canBeNone=True),
    description=_(
u"""Size of the blocks of SST data to cache in memory, in the t
direction (time). The size is given as the number of cells.

If this parameter is 0, no blocks will be cached in memory."""),
    arcGISDisplayName=_(u'In-memory cache block size, in T direction'),
    arcGISCategory=_(u'Performance tuning options'))

AddResultMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'updatedPoints',
    typeMetadata=ArcGISFeatureLayerTypeMetadata(),
    description=_(u'Updated points.'),
    arcGISDisplayName=_(u'Updated points'),
    arcGISParameterDependencies=[u'points'])

# Public method: MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters

AddMethodMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters,
    shortDescription=_(u'Creates rasters indicating the positions of fronts in MODIS Level 3 SST images published by NASA JPL PO.DAAC, using the Cayula and Cornillon (1992) single image edge detection (SIED) algorithm.'),
    longDescription=_MODISL3SSTTimeSeries_LongDescription + _(
u"""

Given a satellite name, temporal resolution, spatial resolution, and
desired SST product, this tool efficiently downloads the corresponding
time series of MODIS Level 3 SST images, executes the Cayula and Cornillon
SIED algorithm to identify fronts, and creates rasters showing the
locations of the fronts.

This tool is complicated and has a lot of parameters. The complex
dynamics of the ocean, the presence of clouds, the difficulty of
detecting and masking them, and the inability to resolve fine-scale
thermal features using the spatial resolution of MODIS Level 3
products make it hard to obtain perfect results. For the best chance
of success, please read all of the documentation carefully.

We configured the default parameters of this tool to balance a number
of tradeoffs. These defaults are different than those used in Cayula
and Cornillon's original study. For example, the original study used a
32x32 moving window while ours is somewhat smaller because we found
that the algorithm appeared to detect more fronts in MODIS images when
a smaller window was used. This is probably due to the fact that
Cayula and Cornillon's original data had a spatial resolution of about
1 km, while MODIS Level 3 has a resolution of about 4.6 km. The
theoretical basis of the algorithm requires that only one front appear
in the window; the 32x32 window appeared to be too large to work well
with MODIS's coarser resolution, often enclosing multiple fronts.

We believe our default values work better for the MODIS Level 3 images
than Cayula and Cornillon's original values, but we have not attempted
to formally validate the results. Please keep that in mind when
applying this tool in your own studies.

Note that the Front Cleanup Options, which are enabled by default,
require that MATLAB 2007b or the MATLAB Component Runtime (MCR) 7.7 be
installed. You can download a free copy of the MCR 7.7 from
http://code.nicholas.duke.edu/projects/mget/wiki/MCR. If you do not
wish to install MATLAB or the MCR, you must disable the Front Cleanup
Options.

**The SIED Algorithm**

""") + _SIEDDescription +
"""
""" + _MODISReferences +
"""

""" + _SIEDReferences,
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Find Cayula-Cornillon Fronts in PO.DAAC MODIS L3 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\MODIS Global Level 3 Mapped SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cls', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'cls')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'satellite', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'satellite')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'temporalResolution', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'temporalResolution')
MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters.__doc__.Obj.GetArgumentByName(u'temporalResolution').Description += _(
u"""

The choice of an appropriate temporal resolution is critical to
successful front detection. In general, front detection works best on
instantaneous images of SST, where temperature gradients are the
sharpest. For best results, choose daily temporal resolution.

You may be tempted to try a lower temporal resolution, such as 8-day
or monthly, because there are fewer clouds. The problem is that the
SST gradients are not as sharp because SST has been averaged over
several images. If the fronts have moved around over the duration of
those images, as is often the case, the SST gradient will be smoothed
out and harder to detect.

Rather than choosing a lower temporal resolution, consider using daily
resolution and then compositing the resulting frontal images. For
example, rather than using 8-day images, find the fronts in the 8
daily images and then add them together. This is similar to how Cayula
and Cornillon's (1996) multiple-image edge detection (MIED) works.
Also consider using both daytime and nighttime images in your
composites.""")

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialResolution', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'spatialResolution')
MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters.__doc__.Obj.GetArgumentByName(u'spatialResolution').Description += _(
u"""

The default Cayula and Cornillon algorithm parameters are intended to be
used with the 4km MODIS products. If you use the 9km products, you
might achieve better results by adjusting the parameter values,
particularly the Histogram Window Size parameter.""")

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'geophysicalParameter', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'geophysicalParameter')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'minPopMeanDifference',
    typeMetadata=FloatTypeMetadata(mustBeGreaterThan=0.),
    description=_(
u"""Minimum difference, in degrees C, between the mean temperatures of
two adjacent populations of pixels for a front to be detected between
those two populations.

The Cayula and Cornillon algorithm passes a moving window over the
image, checking each window for a bimodal distribution in the
temperatures of the pixels within it. When the algorithm detects a
bimodal distribution, it computes the mean temperatures of the two
populations and compares the difference between the means to this
threshold. If the difference is less than this threshold, the
algorithm concludes there is no front present and moves on to the next
window.

You can use this parameter to eliminate weak fronts by selecting a
value that corresponds to a desired minimum mean temperature
difference. Larger values will detect fewer fronts; smaller values
will detect more fronts. However, bear in mind that
`this document <ftp://podaac-ftp.jpl.nasa.gov/allData/modis/docs/MODIS_SST_Guide_Doc.pdf>`_
MODIS SST data can only be considered accurate to +/- 0.4 degrees C.
We do not advise thresholds below that value."""),
    arcGISDisplayName=_(u'Front detection threshold'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWorkspace',
    typeMetadata=ArcGISWorkspaceTypeMetadata(createParentDirectories=True),
    description=_(
u"""Directory or geodatabase to receive the rasters.

Unless you have a specific reason to store the rasters in a
geodatabase, we recommend you store them in a directory because it
will be much faster and allows the rasters to be organized in a tree.
If you do store the rasters in a geodatabase, you must change the
Raster Name Expressions parameter; the documentation for that
parameter for more information.

The front location images will be written as 8-bit signed integer
rasters. Each pixel can have one of three values:

* NoData - the pixel was never a candidate for containing a front,
  either because it was masked or because because it did not appear in
  any histogram windows that had sufficiently large numbers of
  non-masked pixels to proceed with the histogramming step of the
  Cayula and Cornillon algorithm.

* 0 - The pixel was a candidate for containing a front -- it was not
  masked and it appeared in at least one histogram window with a
  sufficient number of non-masked pixels to proceed with the
  histogramming step -- but it was never marked as a front pixel in
  any of the histogram windows it appeared in.

* 1 - The pixel was a candidate for containing a front and it was marked
  as a front pixel in at least one of the histogram windows it
  appeared in.

You may also instruct this tool to write several kinds of diagnostic
outputs. These can help you understand why a front was or was not
identified at a specific location. Please see the documentation for
the diagnostic outputs for more information.

The front location images described above are constructed by
performing a classification of two of the diagnostic outputs, the
"candidate counts" and "front counts" images, and then applying the
Fill Holes, Thin, and Minimum Front Size options."""),
    arcGISDisplayName=_(u'Output workspace'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'mode', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'mode')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure. When storing
rasters in a directory, the final expression specifies the file name
of the raster and any preceding expressions specify subdirectories.
The extension of the final expression determines the output raster
format: .asc for ArcInfo ASCII Grid, .bmp for BMP, .gif for GIF, .img
for an ERDAS IMAGINE file, .jpg for JPEG, .jp2 for JPEG 2000, .png for
PNG, .tif for GeoTIFF, or no extension for ArcInfo Binary Grid. The
default expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(Satellite)s - name of the satellite, either "aqua" or "terra".

* %(SatelliteCode)s - abbreviation for the satellite, either "A" or
  "T", used in NASA's file naming scheme.

* %(TemporalResolution)s - temporal resolution, either "daily",
  "8day", "monthly", or "annual".

* %(TemporalResolutionCode)s - abbreviation for the temporal
  resolution, either "DAY", "8D", "MO", or "YR", used in NASA's file
  naming scheme.

* %(SpatialResolution)s - spatial resolution, either "4km" or "9km".

* %(SpatialResolutionCode)s - abbreviation for the spatial
  resolution, either "4" or "9", used in NASA's file naming scheme.
  
* %(GeophysicalParameter)s - geophysical parameter used to produce the
  output raster, either "sst", "nsst", or "sst4".

* %(Algorithm)s - algorithm used to estimate the geophysical
  parameter from the raw sensor values, either "sst" for the sst or
  nsst geophysical parameter, or "sst4" for the sst4 geophysical
  parameter. The algorithm name is used in NASA's file naming scheme.

* %(Wavelength)s - an alternative abbreviation for the algorithm used
  to estimate the geophysical parameter, either "11um" for the sst or
  nsst geophysical parameter, or "4um" for the sst4 geophysical
  parameter. The abbreviation refers to the wavelengths of the MODIS
  bands used in the algorithm, where 11um refers to the "thermal IR"
  bands and 4um refers to the "mid-IR" bands. The abbreviation is used
  in NASA's file naming scheme.
  
* %(ImageType)s - type of image represented in the output raster,
  either "floc" (front locations), "ccnt" (candidate counts), "fcnt"
  (front counts), "wsco" (window status codes), or "wsvl" (window
  status values).

* %%Y - four-digit year of the raster.

* %%m - two-digit month of the first day of the raster.

* %%d - two-digit day of the month of the first day of the raster.

* %%j - three-digit day of the year of the first day of the raster.

* %(EndDate)s - date of the last day of the raster in the format
  "YYYYjjj" where YYYY is the four-digit year and jjj is the
  three-digit day of the year. This date is used in NASA's file naming
  scheme for temporal resolutions other than "daily".
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'medianFilterWindowSize', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'medianFilterWindowSize')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'histogramWindowSize', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'histogramWindowSize')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'histogramWindowStride', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'histogramWindowStride')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minPropNonMaskedCells', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'minPropNonMaskedCells')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minPopProp', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'minPopProp')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minTheta', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'minTheta')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minSinglePopCohesion', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'minSinglePopCohesion')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minGlobalPopCohesion', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'minGlobalPopCohesion')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'threads', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'threads')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'fillHoles', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'fillHoles')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'thin', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'thin')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minSize', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'minSize')

CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'qualityLevel', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'qualityLevel')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rotationOffset', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'rotationOffset')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialExtent', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'spatialExtent')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'startDate', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'startDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'endDate', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'endDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'timeout', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'timeout')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'maxRetryTime', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cacheDirectory', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'cacheDirectory')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'calculateStatistics', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'calculateStatistics')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildRAT', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'buildRAT')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildPyramids', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'buildPyramids')

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputCandidateCounts',
    typeMetadata=BooleanTypeMetadata(),
    description=_(
u"""If True, candidate counts diagnostic images will be created.

Candidate counts images contain 16-bit unsigned integers and indicate
how many times each pixel of the corresponding SST image was a
candidate for containing a front, i.e. the number of times it appeared
in a histogram window that had a sufficiently large number of
non-masked pixels to proceed with the histogramming step of the
Cayula and Cornillon algorithm. If the histogram window stride is less
than the window size, successive histogram windows will overlap, and
many pixels will have candidate counts greater than 1. Masked pixels
can never be candidates for containing a front so they will always
have the value 0."""),
    arcGISDisplayName=_(u'Create candidate counts rasters'),
    arcGISCategory=_(u'Optional diagnostic outputs'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputFrontCounts',
    typeMetadata=BooleanTypeMetadata(),
    description=_(
u"""If True, front counts diagnostic images will be created.

Front counts images contain 16-bit unsigned integers and indicate how
many times each pixel of the corresponding SST image was found to
contain a front. The value will range from 0 (it never contained a
front) to the candidate count value for the pixel (it always contained
a front in every histogram window that contained it).

The Front Cleanup Options are not applied to the front counts
image."""),
    arcGISDisplayName=_(u'Create front counts rasters'),
    arcGISCategory=_(u'Optional diagnostic outputs'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWindowStatusCodes',
    typeMetadata=BooleanTypeMetadata(),
    description=_(
u"""If True, diagnostic images of window status codes will be created.

Window status code images contain 8-bit unsigned integers that
indicate the result of the algorithm for the histogram window centered
on the pixel. You can use this image to diagnose why the algorithm did
not find fronts in a particular region of your image.

The algorithm result will be one of these code numbers:

""" + _WindowStatusCodes),
    arcGISDisplayName=_(u'Create window status code rasters'),
    arcGISCategory=_(u'Optional diagnostic outputs'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWindowStatusValues',
    typeMetadata=BooleanTypeMetadata(),
    description=_(
u"""If True, diagnostic images of window status values will be
created.

Window status values images contain 32-bit floating point values
to be used in conjunction with the window status codes when diagnosing
the results of the algorithm. The value of each pixel depends on the
window status code for that pixel:

""" + _WindowStatusValues),
    arcGISDisplayName=_(u'Create window status value rasters'),
    arcGISCategory=_(u'Optional diagnostic outputs'))

CopyResultMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'updatedOutputWorkspace', MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'updatedOutputWorkspace')

# Public method: MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters

AddMethodMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters,
    shortDescription=_(u'Creates rasters indicating the positions of fronts in MODIS Level 3 SST images published by NASA JPL PO.DAAC, using the Canny edge detection algorithm.'),
    longDescription=_(
u"""This tool requires that MATLAB 2007b or the MATLAB Component
Runtime (MCR) 7.7 be installed. You can download a free copy of the
MCR 7.7 from http://code.nicholas.duke.edu/projects/mget/wiki/MCR.

""") + _MODISL3SSTTimeSeries_LongDescription + _(
u"""

Given a satellite name, temporal resolution, spatial resolution, and
desired SST product, this tool efficiently downloads the corresponding
time series of MODIS Level 3 SST images, executes the Canny edge
detection algorithm to identify fronts, and creates rasters showing
the locations of the fronts.

""") + _CannyDescription +
u"""

""" + _MODISReferences,
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Find Canny Fronts in PO.DAAC MODIS L3 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\MODIS Global Level 3 Mapped SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cls', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'cls')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'satellite', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'satellite')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'temporalResolution', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'temporalResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialResolution', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'spatialResolution')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'geophysicalParameter', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'geophysicalParameter')

AddArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'outputWorkspace',
    typeMetadata=ArcGISWorkspaceTypeMetadata(createParentDirectories=True),
    description=_(
u"""Directory or geodatabase to receive the rasters.

Unless you have a specific reason to store the rasters in a
geodatabase, we recommend you store them in a directory because it
will be much faster and allows the rasters to be organized in a tree.
If you do store the rasters in a geodatabase, you must change the
Raster Name Expressions parameter; the documentation for that
parameter for more information.

Each raster will contain 8-bit signed integers with three possible
values:

* NoData - the pixel did not have any data (e.g. it was land or
  cloud).

* 0 - a front was detected at the pixel.

* 1 - a front was not detected at the pixel.
"""),
    arcGISDisplayName=_(u'Output workspace'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'mode', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'mode')

AddArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure with names that
imitate those used by NASA. When storing rasters in a directory, the
final expression specifies the file name of the raster and any
preceding expressions specify subdirectories. The extension of the
final expression determines the output raster format: .asc for ArcInfo
ASCII Grid, .bmp for BMP, .gif for GIF, .img for an ERDAS IMAGINE
file, .jpg for JPEG, .jp2 for JPEG 2000, .png for PNG, .tif for
GeoTIFF, or no extension for ArcInfo Binary Grid. The default
expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(Satellite)s - name of the satellite, either "aqua" or "terra".

* %(SatelliteCode)s - abbreviation for the satellite, either "A" or
  "T", used in NASA's file naming scheme.

* %(TemporalResolution)s - temporal resolution, either "daily",
  "8day", "monthly", or "annual".

* %(TemporalResolutionCode)s - abbreviation for the temporal
  resolution, either "DAY", "8D", "MO", or "YR", used in NASA's file
  naming scheme.

* %(SpatialResolution)s - spatial resolution, either "4km" or "9km".

* %(SpatialResolutionCode)s - abbreviation for the spatial
  resolution, either "4" or "9", used in NASA's file naming scheme.
  
* %(GeophysicalParameter)s - geophysical parameter represented in the
  output raster, either "sst", "nsst", or "sst4".

* %(Algorithm)s - algorithm used to estimate the geophysical
  parameter from the raw sensor values, either "sst" for the sst or
  nsst geophysical parameter, or "sst4" for the sst4 geophysical
  parameter. The algorithm name is used in NASA's file naming scheme.

* %(Wavelength)s - an alternative abbreviation for the algorithm used
  to estimate the geophysical parameter, either "11um" for the sst or
  nsst geophysical parameter, or "4um" for the sst4 geophysical
  parameter. The abbreviation refers to the wavelengths of the MODIS
  bands used in the algorithm, where 11um refers to the "thermal IR"
  bands and 4um refers to the "mid-IR" bands. The abbreviation is used
  in NASA's file naming scheme.

* %%Y - four-digit year of the raster.

* %%m - two-digit month of the first day of the raster.

* %%d - two-digit day of the month of the first day of the raster.

* %%j - three-digit day of the year of the first day of the raster.

* %(EndDate)s - date of the last day of the raster in the format
  "YYYYjjj" where YYYY is the four-digit year and jjj is the
  three-digit day of the year. This date is used in NASA's file naming
  scheme for temporal resolutions other than "daily".
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'highThreshold',
    typeMetadata=FloatTypeMetadata(mustBeGreaterThan=0., canBeNone=True),
    description=CannyEdgeGrid.__init__.__doc__.Obj.GetArgumentByName(u'highThreshold').Description,
    arcGISDisplayName=_(u'High threshold'),
    arcGISCategory=_(u'Canny algorithm parameters'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'lowThreshold',
    typeMetadata=FloatTypeMetadata(mustBeGreaterThan=0., canBeNone=True),
    description=CannyEdgeGrid.__init__.__doc__.Obj.GetArgumentByName(u'lowThreshold').Description,
    arcGISDisplayName=_(u'Low threshold'),
    arcGISCategory=_(u'Canny algorithm parameters'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'sigma',
    typeMetadata=FloatTypeMetadata(mustBeGreaterThan=0., canBeNone=True),
    description=CannyEdgeGrid.__init__.__doc__.Obj.GetArgumentByName(u'sigma').Description,
    arcGISDisplayName=_(u'Sigma'),
    arcGISCategory=_(u'Canny algorithm parameters'))

AddArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'minSize',
    typeMetadata=IntegerTypeMetadata(mustBeGreaterThan=0, canBeNone=True),
    description=_(
u"""Minimum number of pixels an individual edge must occupy for it to
be retained. Edges with fewer number of pixels will be discarded.

For this option, an "edge" is defined as the chain of pixels that are
8-connected (i.e. they are connected horizontally, vertically, or
diagonally). The edge may be any shape, such as long and thin and
branching or short and blob-like.

This option is applied after the Canny algorithm is complete. We
selected 4 as the default value, to remove edges that are so small
they are probably not true fronts."""),
    arcGISDisplayName=_(u'Minimum edge size'),
    arcGISCategory=_(u'Canny algorithm parameters'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.__init__, u'qualityLevel', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'qualityLevel')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rotationOffset', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'rotationOffset')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialExtent', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'spatialExtent')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'startDate', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'startDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'endDate', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'endDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'timeout', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'timeout')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'maxRetryTime', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cacheDirectory', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'cacheDirectory')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'calculateStatistics', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'calculateStatistics')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildRAT', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'buildRAT')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildPyramids', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'buildPyramids')

CopyResultMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'updatedOutputWorkspace', MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'updatedOutputWorkspace')

###############################################################################
# Metadata: _GHRSSTLevel4THREDDSCatalog class
###############################################################################

AddClassMetadata(_GHRSSTLevel4THREDDSCatalog,
    shortDescription=_(u'THREDDSCatalog for the NASA JPL PO.DAAC GHRSST L4 datasets.'),
    longDescription=_(
u"""This class is not intended to be used directly by external
callers. Use GHRSSTLevel4 instead."""))

###############################################################################
# Metadata: _GHRSSTLevel4OPeNDAPURL class
###############################################################################

AddClassMetadata(_GHRSSTLevel4OPeNDAPURL,
    shortDescription=_(u'OPeNDAPURL for a NASA JPL PO.DAAC GHRSST L4 file.'),
    longDescription=_(
u"""This class is not intended to be used directly by external
callers. Use GHRSSTLevel4 instead."""))

###############################################################################
# Metadata: _GHRSSTLevel4OPeNDAPGrid class
###############################################################################

AddClassMetadata(_GHRSSTLevel4OPeNDAPGrid,
    shortDescription=_(u'OPeNDAPGrid for a variable of a NASA JPL PO.DAAC GHRSST L4 file.'),
    longDescription=_(
u"""This class is not intended to be used directly by external
callers. Use GHRSSTLevel4 instead."""))

# Constructor

AddMethodMetadata(_GHRSSTLevel4OPeNDAPGrid.__init__,
    shortDescription=_(u'Constructs a new _GHRSSTLevel4OPeNDAPGrid instance.'))

AddArgumentMetadata(_GHRSSTLevel4OPeNDAPGrid.__init__, u'self',
    typeMetadata=ClassInstanceTypeMetadata(cls=_GHRSSTLevel4OPeNDAPGrid),
    description=_(u'_GHRSSTLevel4OPeNDAPGrid instance.'))

AddArgumentMetadata(_GHRSSTLevel4OPeNDAPGrid.__init__, u'opendapURLObj',
    typeMetadata=ClassInstanceTypeMetadata(cls=_GHRSSTLevel4OPeNDAPURL),
    description=_(u'_GHRSSTLevel4OPeNDAPURL instance that contains us.'))

AddArgumentMetadata(_GHRSSTLevel4OPeNDAPGrid.__init__, u'variableName',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'analysed_sst', u'analysis_error', u'mask', u'sea_ice_fraction']),
    description=_(u'OPeNDAP variable name.'))

AddArgumentMetadata(_GHRSSTLevel4OPeNDAPGrid.__init__, u'variableType',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Grid']),
    description=_(u'OPeNDAP variable type.'))

AddResultMetadata(_GHRSSTLevel4OPeNDAPGrid.__init__, u'grid',
    typeMetadata=ClassInstanceTypeMetadata(cls=_GHRSSTLevel4OPeNDAPGrid),
    description=_(u'_GHRSSTLevel4OPeNDAPGrid instance.'))

###############################################################################
# Metadata: GHRSSTLevel4 class
###############################################################################

_GHRSSTLevel4_LongDescription = _(
u"""The `Group for High-Resolution Sea Surface Temperature (GHRSST) <http://www.ghrsst.org/>`_
provides a new generation of global high-resolution (<10km) SST
products to the operational oceanographic, meteorological, climate and
general scientific community. The overall aim of the GHRSST is to
provide the best quality sea surface temperature data for applications
in short, medium and decadal/climate time scales in the most cost
effective and efficient manner through international collaboration and
scientific innovation.

This %(name)s accesses L4 gap-free gridded SST products published in
near real time by the GHRSST GDAC Global Data Assembly Center (GDAC)
at the `NASA JPL Physical Oceanography Distributed Active Archive Center (PO.DAAC) <http://podaac.jpl.nasa.gov/SeaSurfaceTemperature/GHRSST>`_.
These products provide regional and global daily cloud-free estimates
of SST at spatial resolutions ranging from 0.25 degrees to down 1 km.
To fill in cloudy areas, data from multiple satellite and in-situ
sensors are combined and regions without any data are filled in
various interpolation and modeling techniques.

All products accessed by this %(name)s are published at a daily
timestep. Some products are updated on a continual basis and available
in near real time; others are updated infrequently and are intended
mainly for historical analysis. The temporal extent, spatial
resolution and extent, and sensors and interpolation technique used
vary by product. All products use the WGS 1984 geographic coordinate
system. GHRSST temperatures are published in kelvin. By default, this
%(name)s converts them to degrees Celsius but provides an option to
obtain the original kelvin values.

All products are obtained from PO.DAAC using the OPeNDAP protocol.
This %(name)s supports most but not all of the GHRSST L4 products
hosted by PO.DAAC. If you see a product on PO.DAAC that is not
available with this tool, please contact the MGET development team for
assistance.

Please see
`these instructions <http://podaac.jpl.nasa.gov/CitingPODAAC>`_
on how to acknowledge your use of PO.DAAC products.""")

AddClassMetadata(GHRSSTLevel4,
    shortDescription=_(u'A Grid representing a GHRSST L4 product hosted by NASA JPL PO.DAAC.'),
    longDescription=_GHRSSTLevel4_LongDescription % {u'name': 'class'})

# Constructor

AddMethodMetadata(GHRSSTLevel4.__init__,
    shortDescription=_(u'Constructs a new GHRSSTLevel4 instance.'))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'self',
    typeMetadata=ClassInstanceTypeMetadata(cls=GHRSSTLevel4),
    description=_(u'GHRSSTLevel4 instance.'))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'variableName',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'analysed_sst', u'analysis_error']),
    description=_(
u"""GHRSST variable to access, one of:

* analysed_sst - GHRSST product-specific estimate of SST.

* analysis_error - GHRSST product-specific estimate of of the error in
  the SST estimate.

Please see the product documentation for details about what these
variables mean and how they were calculated. These variable names are
case-sensitive."""),
    arcGISDisplayName=_(u'GHRSST variable'))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'productName',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(
u"""Name of the GHRSST L4 product, e.g. 'MUR-JPL-L4-GLOB-v4.1'. The name is
only used for display purposes."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'rdacCode',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(
u"""The code of the GHRSST Regional Data Access Center (RDAC) that
produced the product. These codes are case-sensitive, typically upper
case, and assigned by GHRSST."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'productType',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(
u"""GHRSST product type code. These codes are case-sensitive, assigned
by GHRSST, and indicate the overall resolution and type of SST
estimated by the product. Typical codes include "LRfnd", "HRfnd",
"UHfnd", "HR1m", and "LRblend"."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'areaCode',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(
u"""The region code of the the product. These codes are
case-sensitive, typically upper case, and assigned by GHRSST. A list
of codes may be seen on the PO.DAAC OPeNDAP server at
http://podaac-opendap.jpl.nasa.gov/opendap/hyrax/allData/ghrsst/data/L4/"""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'areaCode2',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(
u"""An abbreviated region code of the the product. Only used for certain
products that have long areaCodes. Usually set to the same as areaCode."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'gdsVersion',
    typeMetadata=IntegerTypeMetadata(minValue=1),
    description=_(
u"""Version of the GHRSST Data Specification (GDS) used by the
product. At the time of this writing, all products accessed by this
function were at GDS version 1."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'fileVersion',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(
u"""Version of the file. This is a case-sensitive string that is
product specific and assigned by the RDAC that produced the
product."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'productCode',
    typeMetadata=UnicodeStringTypeMetadata(),
    description=_(
u"""The RDAC-assigned product code that appears at the end of the
GHRSST file name."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'productCode2',
    typeMetadata=UnicodeStringTypeMetadata(canBeNone=True),
    description=_(
u"""The RDAC-assigned product code that appears in the PO.DAAC OPeNDAP
catalog structure below the RDAC level. For example, in the URL
http://podaac-opendap.jpl.nasa.gov/opendap/hyrax/allData/ghrsst/data/L4/GLOB/REMSS/mw_ir_OI/2012/001/20120103-REMSS-L4HRfnd-GLOB-v01-fv03-mw_ir_rt_OI.nc.gz,
the productCode is "mw_ir_rt_OI" while the productCode2 is "mw_ir_OI".

Typically the productCode2 is the same as the same as the productCode,
in which case you should omit the productCode2 (leave it None). Only a
few products have a different productCode2."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'applyMask',
    typeMetadata=BooleanTypeMetadata(),
    description=_(
u"""If True, temperature values will be masked where the GHRSST mask
variable is some value other than 1, which indicates sea-ice-free
water. If False (the default), no automatic masking will be done.

Most GHRSST L4 datasets already apply the mask, at least for land,
such that the analysed_sst variable takes the "fill value" for land
pixels. The applyMask parameter is provided so that the datasets that
do not do this (e.g. CMC0.2deg-CMC-L4-GLOB-v2.0) can be made
consistent with those that do."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'customPathParsingExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1, canBeNone=True),
    description=_(

u"""Custom regular expressions for parsing path and filenames on the PO.DAAC
server. One regular expression must be provided per directory level below
http://podaac-opendap.jpl.nasa.gov/opendap/allData/ghrsst/data/GDS2/L4/. 

If the default value None is provided, then default regular expressions will
be selected based on the GDS version.

This parameter was introduced to handle the REMSS GDS 2.0 datasets
(MW_OI-REMSS-L4-GLOB-v4.0 and MW_IR_OI-REMSS-L4-GLOB-v4.0) that used a more
elaborate naming scheme than the other L4 products extant at the time."""))

AddArgumentMetadata(GHRSSTLevel4.__init__, u'convertToCelsius',
    typeMetadata=BooleanTypeMetadata(),
    description=_(
u"""If True (the default), temperature values will be converted from
kelvin to degrees Celsius. If False, temperature values will be in the
original kelvin values."""),
    arcGISDisplayName=_(u'Convert temperatures to Celsius'),
    arcGISCategory=_(u'Additional raster processing options'))

CopyArgumentMetadata(THREDDSCatalog.__init__, u'timeout', GHRSSTLevel4.__init__, u'timeout')
CopyArgumentMetadata(THREDDSCatalog.__init__, u'maxRetryTime', GHRSSTLevel4.__init__, u'maxRetryTime')
CopyArgumentMetadata(THREDDSCatalog.__init__, u'cacheDirectory', GHRSSTLevel4.__init__, u'cacheDirectory')
CopyArgumentMetadata(THREDDSCatalog.__init__, u'cacheCatalogsOnDisk', GHRSSTLevel4.__init__, u'cacheCatalogsOnDisk')

AddResultMetadata(GHRSSTLevel4.__init__, u'grid',
    typeMetadata=ClassInstanceTypeMetadata(cls=GHRSSTLevel4),
    description=_(u'GHRSSTLevel4 instance.'))

# Public method: GHRSSTLevel4.CreateArcGISRasters

AddMethodMetadata(GHRSSTLevel4.CreateArcGISRasters,
    shortDescription=_(u'Creates rasters for a GHRSST L4 product hosted by NASA JPL PO.DAAC.'),
    longDescription=_GHRSSTLevel4_LongDescription % {u'name': 'tool'},
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Create Rasters for GHRSST L4 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\GHRSST L4 SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

AddArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cls',
    typeMetadata=ClassOrClassInstanceTypeMetadata(cls=GHRSSTLevel4),
    description=_(u'GHRSSTLevel4 class or instance.'))

AddArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'product',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=sorted(GHRSSTLevel4._PODAAC_Products.keys())),
    description=_(
u"""GHRSST L4 product to access. Currently, the following products are
supported:

""" + '\n'.join(['* `%s <http://podaac.jpl.nasa.gov/dataset/%s>`_.' % (product, product) for product in sorted(GHRSSTLevel4._PODAAC_Products.keys())]) + """

All products use the WGS 1984 geographic coordinate system and are
published at a daily timestep. Some products are updated on a
continual basis and available in near real time; others are updated
infrequently and are intended mainly for historical analysis. The
temporal extent, spatial resolution and extent, and sensors and
interpolation technique used vary by product. Please see the products'
documentation for details.

This tool supports most but not all of the GHRSST L4 products hosted
by PO.DAAC. If you see a product on PO.DAAC that is not available with
this tool, please contact the MGET development team for
assistance."""),
    arcGISDisplayName=_(u'GHRSST L4 product'))

CopyArgumentMetadata(GHRSSTLevel4.__init__, u'variableName', GHRSSTLevel4.CreateArcGISRasters, u'variableName')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'outputWorkspace', GHRSSTLevel4.CreateArcGISRasters, u'outputWorkspace')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'mode', GHRSSTLevel4.CreateArcGISRasters, u'mode')

AddArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure with names that
imitate those used by NASA. When storing rasters in a directory, the
final expression specifies the file name of the raster and any
preceding expressions specify subdirectories. The extension of the
final expression determines the output raster format: .asc for ArcInfo
ASCII Grid, .bmp for BMP, .gif for GIF, .img for an ERDAS IMAGINE
file, .jpg for JPEG, .jp2 for JPEG 2000, .png for PNG, .tif for
GeoTIFF, or no extension for ArcInfo Binary Grid. The default
expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(AreaCode)s - GHRSST abbreviation indicating the spatial extent of
  the product, such as "GLOB" for global extent or "AUS" for
  Australia.

* %(AreaCode2)s - usually the same as AreaCode but for some products a
  further abbreviated version.

* %(RDACCode)s - GHRSST abbreviation for the Regonal Data Assembly
  Center (RDAC) that produced the product.

* %(ProductCode)s - name of the product, assigned by the RDAC.

* %(ProductCode2)s - usually the same as ProductCode but for some
  products a slightly different version. Assigned by the RDAC.

* %(VariableName)s - GHRSST variable represented in the output raster,
  either "analysed_sst" or "analysis_error".

* %(ProductType)s - GHRSST product type code that indicates the
  overall resolution and type of SST estimated by the product. For
  example, "UHfnd" indicates ultra-high resolution (< 5 km spatial
  resolution) foundational SST. Please see the GHRSST documentation
  for the formal meanings of these codes.

* %(GDSVersion)02i - version of the GHRSST Data Specification (GDS)
  used by the product. At the time of this writing, all products
  accessed by this function were at GDS version 1, which will be
  rendered as 01 by the expression %(GDSVersion)02i.

* %(FileVersion)s - version of the data file itself. This is
  product-specific and assigned by the RDAC that produced the product.

* %%Y - four-digit year of the raster.

* %%m - two-digit month of the first day of the raster.

* %%d - two-digit day of the month of the first day of the raster.

* %%j - three-digit day of the year of the first day of the raster.
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rasterCatalog', GHRSSTLevel4.CreateArcGISRasters, u'rasterCatalog')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'rotationOffset', GHRSSTLevel4.CreateArcGISRasters, u'rotationOffset')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'spatialExtent', GHRSSTLevel4.CreateArcGISRasters, u'spatialExtent')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'startDate', GHRSSTLevel4.CreateArcGISRasters, u'startDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'endDate', GHRSSTLevel4.CreateArcGISRasters, u'endDate')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'timeout', GHRSSTLevel4.CreateArcGISRasters, u'timeout')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'maxRetryTime', GHRSSTLevel4.CreateArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'cacheDirectory', GHRSSTLevel4.CreateArcGISRasters, u'cacheDirectory')

CopyArgumentMetadata(GHRSSTLevel4.__init__, u'convertToCelsius', GHRSSTLevel4.CreateArcGISRasters, u'convertToCelsius')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'useUnscaledData', GHRSSTLevel4.CreateArcGISRasters, u'useUnscaledData')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'calculateStatistics', GHRSSTLevel4.CreateArcGISRasters, u'calculateStatistics')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildRAT', GHRSSTLevel4.CreateArcGISRasters, u'buildRAT')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateArcGISRasters, u'buildPyramids', GHRSSTLevel4.CreateArcGISRasters, u'buildPyramids')

AddResultMetadata(GHRSSTLevel4.CreateArcGISRasters, u'updatedOutputWorkspace',
    typeMetadata=ArcGISWorkspaceTypeMetadata(),
    description=_(u'Updated output workspace.'),
    arcGISDisplayName=_(u'Updated output workspace'),
    arcGISParameterDependencies=[u'outputWorkspace'])

# Public method: GHRSSTLevel4.CreateClimatologicalArcGISRasters

AddMethodMetadata(GHRSSTLevel4.CreateClimatologicalArcGISRasters,
    shortDescription=_(u'Creates climatological rasters for a GHRSST L4 product hosted by NASA JPL PO.DAAC.'),
    longDescription=_(
u""" This tool produces rasters showing the climatological average
value (or other statistic) of a time series of GHRSST L4 SST images.
Given a desired GHRSST L4 product, a statistic, and a climatological
bin definition, this tool efficiently downloads the images, classifies
them into bins, and produces a single raster for each bin. Each cell
of the raster is produced by calculating the statistic on the values
of that cell extracted from all of the rasters in the bin.

""" + _GHRSSTLevel4_LongDescription) % {u'name': 'tool'},
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Create Climatological Rasters for GHRSST L4 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\GHRSST L4 SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cls', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'cls')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'product', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'product')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'variableName', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'variableName')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'statistic', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'statistic')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'binType', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'binType')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'outputWorkspace', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'outputWorkspace')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'mode', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'mode')

AddArgumentMetadata(GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure. When storing
rasters in a directory, the final expression specifies the file name
of the raster and any preceding expressions specify subdirectories.
The extension of the final expression determines the output raster
format: .asc for ArcInfo ASCII Grid, .bmp for BMP, .gif for GIF, .img
for an ERDAS IMAGINE file, .jpg for JPEG, .jp2 for JPEG 2000, .png for
PNG, .tif for GeoTIFF, or no extension for ArcInfo Binary Grid. The
default expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(AreaCode)s - GHRSST abbreviation indicating the spatial extent of
  the product, such as "GLOB" for global extent or "AUS" for
  Australia.

* %(AreaCode2)s - usually the same as AreaCode but for some products a
  further abbreviated version.

* %(RDACCode)s - GHRSST abbreviation for the Regonal Data Assembly
  Center (RDAC) that produced the product.

* %(ProductCode)s - name of the product, assigned by the RDAC.

* %(ProductCode2)s - usually the same as ProductCode but for some
  products a slightly different version. Assigned by the RDAC.

* %(VariableName)s - GHRSST variable represented in the output raster,
  either "analysed_sst" or "analysis_error".

* %(ProductType)s - GHRSST product type code that indicates the
  overall resolution and type of SST estimated by the product. For
  example, "UHfnd" indicates ultra-high resolution (< 5 km spatial
  resolution) foundational SST. Please see the GHRSST documentation
  for the formal meanings of these codes.

* %(GDSVersion)02i - version of the GHRSST Data Specification (GDS)
  used by the product. At the time of this writing, all products
  accessed by this function were at GDS version 1, which will be
  rendered as 01 by the expression %(GDSVersion)02i.

* %(FileVersion)s - version of the data file itself. This is
  product-specific and assigned by the RDAC that produced the product.

* %(ClimatologyBinType)s - type of the climatology bin, either "Daily"
  if 1-day bins, "Xday" if multi-day bins (X is replaced by the
  duration), "Monthly" if 1-month bins, "Xmonth" if multi-month bins,
  or "Cumulative". If an ENSO bin type is used, "ENSO\\_" will be
  prepended to those strings (e.g. "ENSO_Daily", "ENSO_Monthly").

* %(ClimatologyBinName)s - name of the climatology bin corresponding
  represented by the output raster, either "dayXXX" for 1-day bins
  (XXX is replaced by the day of the year), "daysXXXtoYYY" for
  multi-day bins (XXX is replaced by the first day of the bin, YYY is
  replaced by the last day), "monthXX" for 1-month bins (XX is
  replaced by the month), "monthXXtoYY" (XX is replaced by the first
  month of the bin, YY by the last month), or "cumulative". If an ENSO
  bin type is used, "neutral\\_", "ElNino\\_", and "LaNina\\_" will be
  prepended to those strings for each of the three ENSO phased rasters
  (e.g. "neutral_cumulative", "ElNino_cumulative", and
  "LaNina_cumulative" when "ENSO Cumulative" bins are requested).

* %(Statistic)s - statistic that was calculated, in lowercase and with
  spaces replaced by underscores; one of: "count", "maximum", "mean",
  "minimum", "range", "standard_deviation", "Sum".

If the Bin Type is "Daily", the following additional codes are
available:

* %(FirstDay)i - first day of the year of the climatology bin
  represented by the output raster.

* %(LastDay)i - last day of the year of the climatology bin
  represented by the output raster. For 1-day climatologies, this will
  be the same as %(FirstDay)i.

If the Bin Type is "Monthly", the following additional codes are
available:

* %(FirstMonth)i - first month of the climatology bin represented by
  the output raster.

* %(DayOfFirstMonth)i - first day of the first month of the
  climatology bin represented by the output raster.

* %(LastMonth)i - last month of the climatology bin represented by
  the output raster.

* %(DayOfLastMonth)i - last day of the last month of the climatology
  bin represented by the output raster.

Note that the additional codes are integers and may be formatted using
"printf"-style formatting codes. For example, to format the FirstDay
as a three-digit number with leading zeros::

    %(FirstDay)03i
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'binDuration', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'binDuration')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateClimatologicalArcGISRasters, u'startDayOfYear', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'startDayOfYear')

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'rotationOffset', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'rotationOffset')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'spatialExtent', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'spatialExtent')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'startDate', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'startDate')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'endDate', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'endDate')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'timeout', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'timeout')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'maxRetryTime', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cacheDirectory', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'cacheDirectory')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'convertToCelsius', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'convertToCelsius')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'calculateStatistics', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'calculateStatistics')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'buildPyramids', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'buildPyramids')

CopyResultMetadata(GHRSSTLevel4.CreateArcGISRasters, u'updatedOutputWorkspace', GHRSSTLevel4.CreateClimatologicalArcGISRasters, u'updatedOutputWorkspace')

# Public method: GHRSSTLevel4.InterpolateAtArcGISPoints

AddMethodMetadata(GHRSSTLevel4.InterpolateAtArcGISPoints,
    shortDescription=_(u'Interpolates values of a GHRSST L4 product hosted by NASA JPL PO.DAAC at points.'),
    longDescription=_(
u""" Given a desired GHRSST L4 product, this tool interpolates the
value of that product at the given points. This tool performs the same
basic operation as the ArcGIS Spatial Analyst's Extract Values to
Points tool, but it reads the MODIS data directly from NASA's servers
rather than reading rasters stored on your machine.

""" + _GHRSSTLevel4_LongDescription) % {u'name': 'tool'},
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Interpolate GHRSST L4 SST at Points'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\GHRSST L4 SST'),
    dependencies=[ArcGISDependency(9, 1, requiresCOMInstantiation=True), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cls', GHRSSTLevel4.InterpolateAtArcGISPoints, u'cls')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'product', GHRSSTLevel4.InterpolateAtArcGISPoints, u'product')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'variableName', GHRSSTLevel4.InterpolateAtArcGISPoints, u'variableName')

CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'points', GHRSSTLevel4.InterpolateAtArcGISPoints, u'points')
CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'valueField', GHRSSTLevel4.InterpolateAtArcGISPoints, u'valueField')
CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'tField', GHRSSTLevel4.InterpolateAtArcGISPoints, u'tField')

AddArgumentMetadata(GHRSSTLevel4.InterpolateAtArcGISPoints, u'method',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Nearest', u'Linear'], makeLowercase=True),
    description=_(
u"""Interpolation method to use, one of:

* Nearest - nearest neighbor interpolation. The interpolated value
  will simply be the value of the cell that contains the point. This
  is the default.

* Linear - linear interpolation (also known as trilinear
  interpolation). This method averages the values of the eight nearest
  cells in the x, y, and time dimensions, weighting the contribution
  of each cell by the area of it that would be covered by a
  hypothetical cell centered on the point being interpolated. If the
  cell containing the point contains NoData, the result is NoData. If
  any of the other seven cells contain NoData, they are omitted from
  the average, and the result is based on the weighted average of the
  cells that do contain data. This is the same algorithm implemented
  by the ArcGIS Spatial Analyst's Extract Values to Points tool.
"""),
    arcGISDisplayName=_(u'Interpolation method'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'where', GHRSSTLevel4.InterpolateAtArcGISPoints, u'where')
CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'noDataValue', GHRSSTLevel4.InterpolateAtArcGISPoints, u'noDataValue')

AddArgumentMetadata(GHRSSTLevel4.InterpolateAtArcGISPoints, u'convertToCelsius',
    typeMetadata=BooleanTypeMetadata(),
    description=_(
u"""If True (the default), temperature values will be converted from
kelvin to degrees Celsius. If False, temperature values will be in the
original kelvin values."""),
    arcGISDisplayName=_(u'Convert temperatures to Celsius'),
    arcGISCategory=_(u'Interpolation options'))

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'timeout', GHRSSTLevel4.InterpolateAtArcGISPoints, u'timeout')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'maxRetryTime', GHRSSTLevel4.InterpolateAtArcGISPoints, u'maxRetryTime')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cacheDirectory', GHRSSTLevel4.InterpolateAtArcGISPoints, u'cacheDirectory')

CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'orderByFields', GHRSSTLevel4.InterpolateAtArcGISPoints, u'orderByFields')

AddArgumentMetadata(GHRSSTLevel4.InterpolateAtArcGISPoints, u'numBlocksToCacheInMemory',
    typeMetadata=IntegerTypeMetadata(minValue=0, canBeNone=True),
    description=_(
u"""Maximum number of blocks of SST data to cache in memory.

To minimize the number of times that the disk or network must be
accessed, this tool employs a simple caching strategy, in addition to
disk caching described by the Cache Directory parameter. When it
processes the first point, it reads a square block of cells centered
on that point and caches it in memory. When it processes the second
and subsequent points, it first checks whether the cells needed for
that point are contained by the block cached in memory. If so, it
processes that point using the in-memory block, rather than reading
from disk or the network again. If not, it reads another square block
centered on that point and adds it to the cache.

The tool processes the remaining points, adding additional blocks to
the cache, as needed. To prevent the cache from exhausing all memory,
it is only permitted to grow to the size specified by this parameter.
When the cache is full but a new block is needed, the oldest block is
discarded to make room for the newest block.

The maximum size of the cache in bytes may be calculated by
multiplying this parameter by the block size parameters and then by 2.
For example, if this parameter is 256 and the blocks are x=64 by y=64
by t=1, the maximum size of the cache is 2097152 bytes (2 MB).

If this parameter is 0, no blocks will be cached in memory."""),
    arcGISDisplayName=_(u'Number of blocks of data to cache in memory'),
    arcGISCategory=_(u'Performance tuning options'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'xBlockSize', GHRSSTLevel4.InterpolateAtArcGISPoints, u'xBlockSize')
CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'yBlockSize', GHRSSTLevel4.InterpolateAtArcGISPoints, u'yBlockSize')
CopyArgumentMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'tBlockSize', GHRSSTLevel4.InterpolateAtArcGISPoints, u'tBlockSize')

CopyResultMetadata(MODISL3SSTTimeSeries.InterpolateAtArcGISPoints, u'updatedPoints', GHRSSTLevel4.InterpolateAtArcGISPoints, u'updatedPoints')

# Public method: GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters

AddMethodMetadata(GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters,
    shortDescription=_(u'Creates rasters indicating the positions of fronts in GHRSST L4 SST images hosted by NASA JPL PO.DAAC, using the Cayula and Cornillon (1992) single image edge detection (SIED) algorithm.'),
    longDescription=_GHRSSTLevel4_LongDescription + _(
u"""

Given a desired GHRSST L4 product, this tool efficiently downloads the
a time series of SST images, executes the Cayula and Cornillon SIED
algorithm to identify fronts, and creates rasters showing the
locations of the fronts.

This tool is complicated and has a lot of parameters. The complex
dynamics of the ocean and the presence of clouds make it hard to
obtain perfect results, even with the sophisticated algorithms GHRSST
products employ to guess values for cloudy pixels. For the best chance
of success, please read all of the documentation carefully.

In our experience, the SIED front detection algorithm is somewhat
sensitive to the spatial resolution of the SST data it is applied to
and will perform better if several parameters are adjusted. We
configured the default parameter values of this tool to those used in
Cayula and Cornillon's original study. That study was applied to AVHRR
images collected by NOAA polar orbiters. The images had an effective
resolution between 1 and 1.5 km. Therefore the default parameter
values for this tool are appropriate for GHRSST products with similar
resolution (e.g. EUR-L4UHRfnd-MED-ODYSSEA, EUR-L4UHRfnd-NWE-ODYSSEA,
JPL-L4UHfnd-GLOB-MUR, JPL_OUROCEAN-L4UHfnd-GLOB-G1SST).

To apply this tool to higher resolution GHRSST L4 products, we suggest
reducing the Histogram Window Size to a smaller value such as 20. If
you do that, you must also reduce the Minimum Single-Population
Cohesion and Minimum Global-Population Cohesion parameters, as
described in the documentation for those parameters. FOr a Histogram
Window Size fo 20, those two parameters should be set to 0.88 and 0.9,
respectively.

Note that the Front Cleanup Options, which are enabled by default,
require that MATLAB 2007b or the MATLAB Component Runtime (MCR) 7.7 be
installed. You can download a free copy of the MCR 7.7 from
http://code.nicholas.duke.edu/projects/mget/wiki/MCR. If you do not
wish to install MATLAB or the MCR, you must disable the Front Cleanup
Options.

**The SIED Algorithm**

""" + _SIEDDescription +
"""

""" + _SIEDReferences) % {u'name': 'tool'},
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Find Cayula-Cornillon Fronts in GHRSST L4 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\GHRSST L4 SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cls', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'cls')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'product', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'product')

AddArgumentMetadata(GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'minPopMeanDifference',
    typeMetadata=FloatTypeMetadata(mustBeGreaterThan=0.),
    description=_(
u"""Minimum difference, in degrees C, between the mean temperatures of
two adjacent populations of pixels for a front to be detected between
those two populations.

The Cayula and Cornillon algorithm passes a moving window over the
image, checking each window for a bimodal distribution in the
temperatures of the pixels within it. When the algorithm detects a
bimodal distribution, it computes the mean temperatures of the two
populations and compares the difference between the means to this
threshold. If the difference is less than this threshold, the
algorithm concludes there is no front present and moves on to the next
window.

You can use this parameter to eliminate weak fronts by selecting a
value that corresponds to a desired minimum mean temperature
difference. Larger values will detect fewer fronts; smaller values
will detect more fronts. However, bear in mind that temperature
estimates derived from most satellite remote sensors are only
considered accurate to within 0.3 to 0.5 degrees C. We do not advise
thresholds below 0.3 to 0.5 degrees C."""),
    arcGISDisplayName=_(u'Front detection threshold'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWorkspace', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWorkspace')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'mode', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'mode')

AddArgumentMetadata(GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure. When storing
rasters in a directory, the final expression specifies the file name
of the raster and any preceding expressions specify subdirectories.
The extension of the final expression determines the output raster
format: .asc for ArcInfo ASCII Grid, .bmp for BMP, .gif for GIF, .img
for an ERDAS IMAGINE file, .jpg for JPEG, .jp2 for JPEG 2000, .png for
PNG, .tif for GeoTIFF, or no extension for ArcInfo Binary Grid. The
default expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(AreaCode)s - GHRSST abbreviation indicating the spatial extent of
  the product, such as "GLOB" for global extent or "AUS" for
  Australia.

* %(AreaCode2)s - usually the same as AreaCode but for some products a
  further abbreviated version.

* %(RDACCode)s - GHRSST abbreviation for the Regonal Data Assembly
  Center (RDAC) that produced the product.

* %(ProductCode)s - name of the product, assigned by the RDAC.

* %(ProductCode2)s - usually the same as ProductCode but for some
  products a slightly different version. Assigned by the RDAC.

* %(VariableName)s - GHRSST variable represented in the output raster;
  always "analysed_sst".

* %(ProductType)s - GHRSST product type code that indicates the
  overall resolution and type of SST estimated by the product. For
  example, "UHfnd" indicates ultra-high resolution (< 5 km spatial
  resolution) foundational SST. Please see the GHRSST documentation
  for the formal meanings of these codes.

* %(GDSVersion)02i - version of the GHRSST Data Specification (GDS)
  used by the product. At the time of this writing, all products
  accessed by this function were at GDS version 1, which will be
  rendered as 01 by the expression %(GDSVersion)02i.

* %(FileVersion)s - version of the data file itself. This is
  product-specific and assigned by the RDAC that produced the product.
  
* %(ImageType)s - type of image represented in the output raster,
  either "floc" (front locations), "ccnt" (candidate counts), "fcnt"
  (front counts), "wsco" (window status codes), or "wsvl" (window
  status values).

* %%Y - four-digit year of the raster.

* %%m - two-digit month of the first day of the raster.

* %%d - two-digit day of the month of the first day of the raster.

* %%j - three-digit day of the year of the first day of the raster.
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'medianFilterWindowSize', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'medianFilterWindowSize')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'histogramWindowSize', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'histogramWindowSize')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'histogramWindowStride', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'histogramWindowStride')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minPropNonMaskedCells', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'minPropNonMaskedCells')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minPopProp', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'minPopProp')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minTheta', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'minTheta')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minSinglePopCohesion', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'minSinglePopCohesion')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minGlobalPopCohesion', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'minGlobalPopCohesion')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'threads', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'threads')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'fillHoles', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'fillHoles')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'thin', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'thin')
CopyArgumentMetadata(CayulaCornillonEdgeDetection.DetectEdgesInSingleImage, u'minSize', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'minSize')

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'rotationOffset', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'rotationOffset')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'spatialExtent', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'spatialExtent')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'startDate', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'startDate')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'endDate', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'endDate')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'timeout', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'timeout')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'maxRetryTime', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cacheDirectory', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'cacheDirectory')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'calculateStatistics', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'calculateStatistics')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'buildRAT', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'buildRAT')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'buildPyramids', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'buildPyramids')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputCandidateCounts', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputCandidateCounts')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputFrontCounts', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputFrontCounts')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWindowStatusCodes', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWindowStatusCodes')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWindowStatusValues', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'outputWindowStatusValues')

CopyResultMetadata(GHRSSTLevel4.CreateArcGISRasters, u'updatedOutputWorkspace', GHRSSTLevel4.CreateCayulaCornillonFrontsAsArcGISRasters, u'updatedOutputWorkspace')

# Public method: GHRSSTLevel4.CannyEdgesAsArcGISRasters

AddMethodMetadata(GHRSSTLevel4.CannyEdgesAsArcGISRasters,
    shortDescription=_(u'Creates rasters indicating the positions of fronts in GHRSST L4 SST images hosted by NASA JPL PO.DAAC, using the Canny edge detection algorithm.'),
    longDescription=_(
u"""This tool requires that MATLAB 2007b or the MATLAB Component
Runtime (MCR) 7.7 be installed. You can download a free copy of the
MCR 7.7 from http://code.nicholas.duke.edu/projects/mget/wiki/MCR.

""") + _GHRSSTLevel4_LongDescription + _(
u"""

Given a desired GHRSST L4 product, this tool efficiently downloads the
a time series of SST images, executes the Canny edge detection
algorithm to identify fronts, and creates rasters showing the
locations of the fronts.

""" + _CannyDescription),
    isExposedToPythonCallers=True,
    isExposedByCOM=True,
    isExposedAsArcGISTool=True,
    arcGISDisplayName=_(u'Find Canny Fronts in GHRSST L4 SST'),
    arcGISToolCategory=_(u'Data Products\\NASA JPL PO.DAAC\\GHRSST L4 SST'),
    dependencies=[ArcGISDependency(9, 1), PythonAggregatedModuleDependency('numpy')])

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cls', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'cls')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'product', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'product')

CopyArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'outputWorkspace', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'outputWorkspace')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'mode', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'mode')

AddArgumentMetadata(GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'rasterNameExpressions',
    typeMetadata=ListTypeMetadata(elementType=UnicodeStringTypeMetadata(), minLength=1),
    description=_(
u"""List of expressions specifying how the output rasters should be
named.

The default expression assumes you are storing rasters in a file
system directory and creates them in a tree structure. When storing
rasters in a directory, the final expression specifies the file name
of the raster and any preceding expressions specify subdirectories.
The extension of the final expression determines the output raster
format: .asc for ArcInfo ASCII Grid, .bmp for BMP, .gif for GIF, .img
for an ERDAS IMAGINE file, .jpg for JPEG, .jp2 for JPEG 2000, .png for
PNG, .tif for GeoTIFF, or no extension for ArcInfo Binary Grid. The
default expression uses .img.

When storing rasters in a geodatabase, you should provide only one
expression. That expression specifies the raster's name.

Each expression may contain any sequence of characters permitted by
the output workspace. Each expression may optionally contain one or
more of the following case-sensitive codes. The tool replaces the
codes with appropriate values when creating each raster:

* %(AreaCode)s - GHRSST abbreviation indicating the spatial extent of
  the product, such as "GLOB" for global extent or "AUS" for
  Australia.

* %(AreaCode2)s - usually the same as AreaCode but for some products a
  further abbreviated version.

* %(RDACCode)s - GHRSST abbreviation for the Regonal Data Assembly
  Center (RDAC) that produced the product.

* %(ProductCode)s - name of the product, assigned by the RDAC.

* %(ProductCode2)s - usually the same as ProductCode but for some
  products a slightly different version. Assigned by the RDAC.

* %(VariableName)s - GHRSST variable represented in the output raster;
  always "analysed_sst".

* %(ProductType)s - GHRSST product type code that indicates the
  overall resolution and type of SST estimated by the product. For
  example, "UHfnd" indicates ultra-high resolution (< 5 km spatial
  resolution) foundational SST. Please see the GHRSST documentation
  for the formal meanings of these codes.

* %(GDSVersion)02i - version of the GHRSST Data Specification (GDS)
  used by the product. At the time of this writing, all products
  accessed by this function were at GDS version 1, which will be
  rendered as 01 by the expression %(GDSVersion)02i.

* %(FileVersion)s - version of the data file itself. This is
  product-specific and assigned by the RDAC that produced the product.

* %%Y - four-digit year of the raster.

* %%m - two-digit month of the first day of the raster.

* %%d - two-digit day of the month of the first day of the raster.

* %%j - three-digit day of the year of the first day of the raster.
"""),
    arcGISDisplayName=_(u'Raster name expressions'))

CopyArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'highThreshold', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'highThreshold')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'lowThreshold', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'lowThreshold')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'sigma', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'sigma')
CopyArgumentMetadata(MODISL3SSTTimeSeries.CannyEdgesAsArcGISRasters, u'minSize', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'minSize')

CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'rotationOffset', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'rotationOffset')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'spatialExtent', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'spatialExtent')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'startDate', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'startDate')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'endDate', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'endDate')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'timeout', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'timeout')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'maxRetryTime', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'maxRetryTime')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'cacheDirectory', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'cacheDirectory')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'calculateStatistics', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'calculateStatistics')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'buildRAT', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'buildRAT')
CopyArgumentMetadata(GHRSSTLevel4.CreateArcGISRasters, u'buildPyramids', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'buildPyramids')

CopyResultMetadata(GHRSSTLevel4.CreateArcGISRasters, u'updatedOutputWorkspace', GHRSSTLevel4.CannyEdgesAsArcGISRasters, u'updatedOutputWorkspace')

###############################################################################
# Metadata: _CCMPWindsL35THREDDSCatalog class
###############################################################################

AddClassMetadata(_CCMPWindsL35THREDDSCatalog,
    shortDescription=_(u'THREDDSCatalog for the NASA JPL PO.DAAC Cross-Calibrated Multi-Platform (CCMP) Ocean Surface Wind Vector L3.5 datasets.'),
    longDescription=_(
u"""This class is not intended to be used directly by external
callers. Use CCMPWindsL35 instead."""))

###############################################################################
# Metadata: _CCMPWindsL35OPeNDAPURL class
###############################################################################

AddClassMetadata(_CCMPWindsL35OPeNDAPURL,
    shortDescription=_(u'OPeNDAPURL for a NASA JPL PO.DAAC Cross-Calibrated Multi-Platform (CCMP) Ocean Surface Wind Vector L3.5 file.'),
    longDescription=_(
u"""This class is not intended to be used directly by external
callers. Use CCMPWindsL35 instead."""))

###############################################################################
# Metadata: _CCMPWindsL35OPeNDAPGrid class
###############################################################################

AddClassMetadata(_CCMPWindsL35OPeNDAPGrid,
    shortDescription=_(u'OPeNDAPGrid for a variable of a NASA JPL PO.DAAC Cross-Calibrated Multi-Platform (CCMP) Ocean Surface Wind Vector L3.5 file.'),
    longDescription=_(
u"""This class is not intended to be used directly by external
callers. Use CCMPWindsL35 instead."""))

# Constructor

AddMethodMetadata(_CCMPWindsL35OPeNDAPGrid.__init__,
    shortDescription=_(u'Constructs a new _CCMPWindsL35OPeNDAPGrid instance.'))

AddArgumentMetadata(_CCMPWindsL35OPeNDAPGrid.__init__, u'self',
    typeMetadata=ClassInstanceTypeMetadata(cls=_CCMPWindsL35OPeNDAPGrid),
    description=_(u'_CCMPWindsL35OPeNDAPGrid instance.'))

AddArgumentMetadata(_CCMPWindsL35OPeNDAPGrid.__init__, u'opendapURLObj',
    typeMetadata=ClassInstanceTypeMetadata(cls=_CCMPWindsL35OPeNDAPURL),
    description=_(u'_CCMPWindsL35OPeNDAPURL instance that contains us.'))

AddArgumentMetadata(_CCMPWindsL35OPeNDAPGrid.__init__, u'variableName',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'uwnd', u'vwnd', u'upstr', u'vpstr', u'wspd', u'nobs']),
    description=_(u'OPeNDAP variable name.'))

AddArgumentMetadata(_CCMPWindsL35OPeNDAPGrid.__init__, u'variableType',
    typeMetadata=UnicodeStringTypeMetadata(allowedValues=[u'Grid']),
    description=_(u'OPeNDAP variable type.'))

AddResultMetadata(_CCMPWindsL35OPeNDAPGrid.__init__, u'grid',
    typeMetadata=ClassInstanceTypeMetadata(cls=_CCMPWindsL35OPeNDAPGrid),
    description=_(u'_CCMPWindsL35OPeNDAPGrid instance.'))

###############################################################################
# Names exported by this module
###############################################################################

__all__ = ['MODISL3SSTTimeSeries',
           'GHRSSTLevel4',
           'CCMPWindsL35']
